from __future__ import annotations

from pathlib import Path

import gdstk
import numpy as np
import orjson
from omegaconf import OmegaConf

from gdsfactory.cell import cell_import_gds
from gdsfactory.component import CellSettings, Component, Info
from gdsfactory.component_reference import ComponentReference
from gdsfactory.config import logger
from gdsfactory.typings import Callable


@cell_import_gds
def import_gds(
    gdspath: str | Path,
    cellname: str | None = None,
    gdsdir: str | Path | None = None,
    read_metadata: bool = False,
    read_metadata_json: bool = False,
    keep_name_short: bool = False,
    unique_names: bool = True,
    max_name_length: int = 250,
    post_process: Callable | None = None,
    **kwargs,
) -> Component:
    """Returns a Component from a GDS file.

    appends $ with a number to the name if the cell name is on CACHE

    Args:
        gdspath: path of GDS file.
        cellname: cell of the name to import. None imports top cell.
        gdsdir: optional GDS directory.
        read_metadata: loads metadata (ports, settings) if it exists in YAML format.
        read_metadata_json: loads metadata (ports, settings) if it exists in JSON format.
        keep_name_short: appends a hash to a shortened component name.
        unique_names: appends $ with a number to the name if the cell name is on CACHE. \
                This avoids name collisions when importing multiple times the same cell name.
        max_name_length: maximum length of the name.
        post_process: function to post process the component after importing.
        kwargs: extra to add to component.info (polarization, wavelength ...).
    """
    gdspath = Path(gdsdir) / Path(gdspath) if gdsdir else Path(gdspath)
    if not gdspath.exists():
        raise FileNotFoundError(f"No file {str(gdspath)!r} found")

    metadata_filepath = gdspath.with_suffix(".yml")
    metadata_json_filepath = gdspath.with_suffix(".json")

    if gdspath.suffix.lower() == ".gds":
        gdsii_lib = gdstk.read_gds(str(gdspath))
    elif gdspath.suffix.lower() == ".oas":
        gdsii_lib = gdstk.read_oas(str(gdspath))
    else:
        raise ValueError(f"gdspath.suffix {gdspath.suffix!r} not .gds or .oas")

    top_level_cells = gdsii_lib.top_level()
    top_cellnames = [c.name for c in top_level_cells]

    if not top_cellnames:
        raise ValueError(f"no top cells found in {str(gdspath)!r}")

    D_list = []
    cell_name_to_component = {}
    cell_to_component = {}

    # create a new Component for each gdstk Cell
    for c in gdsii_lib.cells:
        D = Component()
        D._cell = c
        if not keep_name_short:
            max_name_length = 10000000000000
        D.rename(c.name, cache=unique_names, max_name_length=max_name_length)

        cell_name_to_component[c.name] = D
        cell_to_component[c] = D
        D_list += [D]

    cellnames = list(cell_name_to_component.keys())
    if cellname:
        if cellname not in cell_name_to_component:
            raise ValueError(
                f"cell {cellname!r} is not in file {gdspath} with cells {cellnames}"
            )
    elif len(top_level_cells) == 1:
        cellname = top_level_cells[0].name
    elif len(top_level_cells) > 1:
        raise ValueError(
            f"import_gds() There are multiple top-level cells in {gdspath!r}, "
            f"you must specify `cellname` to select of one of them among {cellnames}"
        )

    # create a new ComponentReference for each gdstk CellReference
    for c, D in cell_to_component.items():
        for e in c.references:
            ref_device = cell_to_component[e.cell]
            ref = ComponentReference(
                component=ref_device,
                origin=e.origin,
                rotation=e.rotation,
                magnification=e.magnification,
                x_reflection=e.x_reflection,
                columns=e.repetition.columns or 1,
                rows=e.repetition.rows or 1,
                spacing=e.repetition.spacing,
                v1=e.repetition.v1,
                v2=e.repetition.v2,
            )
            D._register_reference(ref)
            D._references.append(ref)
            ref._reference = e

    component = cell_name_to_component[cellname]

    if read_metadata and metadata_filepath.exists():
        logger.info(f"Read YAML metadata from {metadata_filepath}")
        metadata = OmegaConf.load(metadata_filepath)

        if "settings" in metadata:
            settings = OmegaConf.to_container(metadata.settings)
            if settings:
                component.settings = CellSettings(**settings)
        if "info" in metadata:
            info = OmegaConf.to_container(metadata.info)
            if info:
                component.info = Info(**info)
        if "function" in metadata:
            component.function_name = metadata.function
        if "module" in metadata:
            component.module = metadata.module

        if "ports" in metadata:
            for port_name, port in metadata.ports.items():
                if port_name not in component.ports:
                    component.add_port(
                        name=port_name,
                        center=np.array(port.center, dtype="float64"),
                        width=port.width,
                        orientation=port.orientation,
                        layer=tuple(port.layer),
                        port_type=port.port_type,
                    )

    if read_metadata_json and metadata_json_filepath.exists():
        logger.info(f"Read JSON metadata from {metadata_json_filepath}")
        metadata = orjson.loads(open(metadata_json_filepath, "rb").read())

        if "settings" in metadata:
            if metadata.get("settings", {}):
                component.settings = CellSettings(**metadata["settings"])

        if "info" in metadata:
            if metadata["info"]:
                component.info = Info(**metadata["info"])
        if "function" in metadata:
            component.function_name = metadata["function"]
        if "module" in metadata:
            component.module = metadata["module"]

        if "ports" in metadata:
            for port_name, port in metadata["ports"].items():
                if port_name not in component.ports:
                    component.add_port(
                        name=port_name,
                        center=np.array(port["center"], dtype="float64"),
                        width=port["width"],
                        orientation=port["orientation"],
                        layer=tuple(port["layer"]),
                        port_type=port["port_type"],
                    )

    for k, v in kwargs.items():
        component.info[k] = v
    component.imported_gds = True
    if post_process:
        post_process(component)
    return component


def import_gds_raw(gdspath, top_cellname: str | None = None):
    if not top_cellname:
        if gdspath.suffix.lower() == ".gds":
            gdsii_lib = gdstk.read_gds(str(gdspath))
        elif gdspath.suffix.lower() == ".oas":
            gdsii_lib = gdstk.read_oas(str(gdspath))
        top_level_cells = gdsii_lib.top_level()
        top_cellnames = [c.name for c in top_level_cells]
        top_cellname = top_cellnames[0]

    cells = gdstk.read_rawcells(gdspath)
    c = Component(name=top_cellname)
    c._cell = cells.pop(top_cellname)
    return c


if __name__ == "__main__":
    import gdsfactory as gf

    # c = gf.Component(name="a" * 250)
    # _ = c << gf.components.mzi()
    c = gf.c.straight()
    gdspath = c.write_gds("a.gds", with_metadata=True)

    # c = import_gds(gdspath)
    c = import_gds("a.gds", read_metadata=True)
    c.show(show_ports=False)

    # gdspath = PATH.gdsdir / "mzi2x2.gds"
    # c = import_gds(gdspath, flatten=True, name="TOP")
    # c.settings = {}
    # print(clean_value_name(c))
    # c = import_gds(gdspath, flatten=False, polarization="te")
    # c = import_gds("/home/jmatres/gdsfactory/gdsfactory/gdsdiff/gds_diff_git.py")
    # print(c.hash_geometry())



from __future__ import annotations

import datetime
import platform

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.text import text
from gdsfactory.typings import LayerSpec, LayerSpecs


@gf.cell
def pixel(
    size: int = 1.0, layer: LayerSpec = "WG", layers: LayerSpecs | None = None
) -> Component:
    c = gf.Component()
    a = size / 2
    layers = layers or [layer]
    for layer in layers:
        c.add_polygon([(a, a), (a, -a), (-a, -a), (-a, a)], layer)
    return c


@gf.cell
def qrcode(
    data: str = "mask01",
    psize: int = 1,
    layer: LayerSpec = "WG",
) -> Component:
    """Returns QRCode."""
    import qrcode

    pix = pixel(size=psize, layer=layer)
    q = qrcode.QRCode()
    q.add_data(data)
    matrix = q.get_matrix()
    c = gf.Component()
    for i, row in enumerate(matrix):
        for j, value in enumerate(row):
            if value:
                pix_ref = pix.ref((i * psize, j * psize))
                c.add(pix_ref)
                c.absorb(pix_ref)
    return c


@gf.cell
def version_stamp(
    labels: tuple[str, ...] = ("demo_label",),
    with_qr_code: bool = False,
    layer: LayerSpec = "WG",
    pixel_size: int = 1,
    version: str | None = None,
    text_size: int = 10,
    layers: LayerSpecs | None = None,
) -> Component:
    """Component with module version and date.

    Args:
        labels: Iterable of labels.
        with_qr_code: Add QR code with timestamp.
        layer: Layer for text.
        pixel_size: Size of the pixel for the QR code.
        version: Version string.
        text_size: Size of the text.
        layers: List of layers for text.
    """
    now = datetime.datetime.now()
    timestamp = f"{now:%Y-%m-%d %H:%M:%S}"
    short_stamp = f"{now:%y.%m.%d.%H.%M.%S}"

    c = gf.Component()
    layers = layers or [layer]

    for layer in layers:
        if with_qr_code:
            data = f"{timestamp}/{platform.node()}"
            q = qrcode(layer=layer, data=data, psize=pixel_size).ref_center()
            c.add(q)
            c.absorb(q)
            x = q.size_info.width * 0.5 + 10

        else:
            x = 0

        txt_params = {"layer": layer, "justify": "left", "size": text_size}
        date = text(
            position=(x, text_size + 2 * pixel_size), text=short_stamp, **txt_params
        ).ref()
        c.add(date)
        c.absorb(date)

        if version:
            t = text(position=(x, 0), text=version, **txt_params).ref()
            c.add(t)
            c.absorb(t)

        for i, line in enumerate(labels):
            t = c << text(
                position=(x, -(i + 1) * (text_size + 2 * pixel_size)),
                text=line,
                **txt_params,
            )
            c.absorb(t)

    return c


if __name__ == "__main__":
    layers = {(1, 0), (2, 0)}
    c = version_stamp(
        pixel_size=4,
        layer="M1",
        with_qr_code=True,
        # text=["b1", "demo"],
        text_size=20,
        layers=layers,
    )
    c.show(show_ports=True)



from functools import partial

import gdsfactory as gf

layer_bbox = (1, 0)


@gf.cell
def sample_add_pads() -> gf.Component:
    """Sample component."""
    c = gf.Component()
    c.add_polygon(
        [[0.0, 25.0], [374.0, 25.0], [374.0, -25.0], [0.0, -25.0]], layer=layer_bbox
    )

    c.add_port(
        name="g1",
        width=100.0,
        cross_section="xs_m3",
        center=(57.0, 25.0),
        orientation=90,
        port_type="electrical",
    )
    c.add_port(
        name="g2",
        width=100.0,
        cross_section="xs_m3",
        center=(57.0, -25.0),
        orientation=-90,
        port_type="electrical",
    )
    c.add_port(
        name="g3",
        width=100.0,
        cross_section="xs_m3",
        center=(317.0, 25.0),
        orientation=90,
        port_type="electrical",
    )
    c.add_port(
        name="g4",
        width=100.0,
        cross_section="xs_m3",
        center=(317.0, -25.0),
        orientation=-90,
        port_type="electrical",
    )
    c.add_port(
        name="o1",
        width=2.0,
        cross_section="xs_sc",
        center=(0.0, 0.0),
        orientation=180,
        port_type="optical",
    )
    c.add_port(
        name="o2",
        width=2.0,
        cross_section="xs_sc",
        center=(374.0, 0.0),
        orientation=0,
        port_type="optical",
    )
    c.add_port(
        name="s1",
        width=60.0,
        cross_section="xs_m3",
        center=(187.0, 25.0),
        orientation=90,
        port_type="electrical",
    )
    c.add_port(
        name="s2",
        width=60.0,
        cross_section="xs_m3",
        center=(187.0, -25.0),
        orientation=-90,
        port_type="electrical",
    )
    return c


def test_sample_add_pads() -> None:
    p = partial(gf.c.pad, size=(100, 100))
    c = sample_add_pads()
    cc = gf.routing.add_pads_top(
        c,
        straight_separation=26,
        pad_spacing=150,
        pad=p,
    )
    assert cc


if __name__ == "__main__":
    p = partial(gf.components.pad, size=(100, 100))
    c = sample_add_pads()
    cc = gf.routing.add_pads_top(
        c,
        straight_separation=26,
        pad_spacing=150,
        pad=p,
    )
    cc.show()



from functools import partial

import numpy as np

import gdsfactory as gf
from gdsfactory.components.cutback_bend import cutback_bend90, cutback_bend180
from gdsfactory.components.cutback_component import cutback_component
from gdsfactory.components.mmi1x2 import mmi1x2
from gdsfactory.components.spiral_inner_io import spiral_inner_io
from gdsfactory.typings import ComponentFactory, CrossSectionSpec

_loss_target = tuple(1 + 1 * i for i in range(3))


def cutback_loss(
    component: ComponentFactory = mmi1x2,
    cutback: ComponentFactory = cutback_component,
    loss: tuple[float, ...] = _loss_target,
    loss_dB: float = 10e-3,
    cols: int | None = 4,
    rows: int | None = None,
    enforce_odd_rows: bool = True,
    decorator: ComponentFactory | None = None,
    **kwargs,
) -> list[gf.Component]:
    """Returns a list of component cutbacks with specified rows and columns to achieve the desired losses.

    Creates a list of components with number of components to achieve a list of specific optical losses.
    The function takes a base component and a cutback component factory as inputs, along with a list of target losses,
    and returns a list of these components with specified rows and columns to achieve the desired losses.

    Args:
        component: component factory.
        cutback: cutback function.
        loss: list of target loss in dB.
        loss_dB: loss per component.
        cols: number of columns.
        rows: number of rows.
        enforce_odd_rows: if True, forces odd number of rows.
        decorator: optional decorator function.

    Keyword Args:
        port1: name of first optical port.
        port2: name of second optical port.
        bend180: ubend.
        mirror: Flips component. Useful when 'o2' is the port that you want to route to.
        mirror1: mirrors first component.
        mirror2: mirrors second component.
        straight_length: length of the straight section between cutbacks.
        straight_length_pair: length of the straight section between each component pair.
        cross_section: specification (CrossSection, string or dict).
        kwargs: component settings.

    """
    loss = np.array(loss)

    if rows and cols:
        raise ValueError("Specify either cols or rows")
    elif rows is None:
        rows_list = np.round(loss / loss_dB / 4 / cols)
        if enforce_odd_rows:
            rows_list = rows_list // 2 * 2 + 1

        components = [
            cutback(component=component, rows=int(rows) or 1, cols=cols, **kwargs)
            for rows in set(rows_list)
        ]
    elif cols is None:
        cols_list = np.round(loss / loss_dB / 4 / rows)
        components = [
            cutback(component=component, rows=rows, cols=int(cols) or 1, **kwargs)
            for cols in set(cols_list)
        ]

    else:
        raise ValueError("Specify either cols or rows")

    if decorator:
        components = [decorator(component) for component in components]
    return components


def cutback_loss_spirals(
    spiral: ComponentFactory = spiral_inner_io,
    loss: tuple[float, ...] = tuple(4 + 3 * i for i in range(3)),
    cross_section: CrossSectionSpec = "xs_sc",
    loss_dB_per_m: float = 300,
    **kwargs,
) -> list[gf.Component]:
    """Returns a list of spirals.

    Args:
        spiral: spiral factory.
        loss: list of target loss in dB.
        cross_section: strip or rib.
        loss_dB_per_m: loss per meter.
        kwargs: additional spiral arguments.
    """
    lengths = [loss_dB / loss_dB_per_m * 1e6 for loss_dB in loss]
    return [
        spiral(length=length, cross_section=cross_section, **kwargs)
        for length in lengths
    ]


cutback_loss_mmi1x2 = partial(cutback_loss, component=mmi1x2, port2="o3", mirror2=True)
cutback_loss_bend90 = partial(
    cutback_loss, component="bend_euler", cutback=cutback_bend90, cols=12
)
cutback_loss_bend180 = partial(
    cutback_loss, component="bend_euler180", cutback=cutback_bend180, cols=12
)


if __name__ == "__main__":
    # c = gf.pack(
    #     cutback_loss_spirals(
    #         decorator=gf.c.add_grating_couplers_with_loopback_fiber_array
    #     )
    # )[0]
    # components = cutback_loss(
    #     component=gf.c.mmi2x2, decorator=gf.routing.add_fiber_array
    # )
    # components = cutback_loss_mmi1x2(decorator=gf.routing.add_fiber_array)
    components = cutback_loss_bend180()
    c = gf.pack(components)[0]
    # c = components[0]
    c.show()



"""Straight waveguide."""
from __future__ import annotations

from collections.abc import Callable

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.cross_section import CrossSectionSpec


@gf.cell
def straight(
    length: float = 10.0,
    npoints: int = 2,
    layer: gf.typings.LayerSpec | None = None,
    width: float | None = None,
    add_pins: bool = True,
    cross_section: CrossSectionSpec = "xs_sc",
    add_bbox: Callable | None = None,
    post_process: Callable | None = None,
) -> Component:
    """Returns a Straight waveguide.

    Args:
        length: straight length (um).
        npoints: number of points.
        layer: layer to use. Defaults to cross_section.layer.
        width: width to use. Defaults to cross_section.width.
        add_pins: add pins to the component.
        cross_section: specification (CrossSection, string or dict).
        add_bbox: function to add bounding box to the component.
        post_process: function to post process the component.

    .. code::

        o1 -------------- o2
                length
    """

    p = gf.path.straight(length=length, npoints=npoints)
    kwargs = dict()
    if layer:
        kwargs["layer"] = layer
    if width:
        kwargs["width"] = width
    x = gf.get_cross_section(cross_section, **kwargs)

    c = Component()
    path = p.extrude(x)
    ref = c << path
    c.add_ports(ref.ports)

    if add_bbox:
        add_bbox(c)
    else:
        x.add_bbox(c, right=0, left=0)
    if add_pins:
        x.add_pins(c)
    c.info["length"] = float(length)
    if len(x.sections) == 0:
        c.info["width"] = x.width
    else:
        c.info["width"] = x.sections[0].width

    c.add_route_info(cross_section=x, length=length)
    c.absorb(ref)
    if post_process:
        post_process(c)
    return c


if __name__ == "__main__":
    import gdsfactory as gf

    # xs = gf.cross_section.pn()
    # xs = xs.mirror()
    # c = straight(cross_section=xs, info=dict(simulation="eme"))
    # print(c.info["simulation"])
    c = gf.Component()
    ref = c << straight(width=3e-3, length=3e-3)
    # ref.xmin = 0
    # ref.ymin = 0
    ref.center = (0, 0)

    c.show()
    # print(c.bbox)
    # c._repr_html_()
    # c.show()
    # c.show(show_ports=True)
    # print(c.bbox)



from __future__ import annotations

from functools import partial

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.port import Port
from gdsfactory.snap import snap_to_grid
from gdsfactory.typings import Callable, CrossSectionSpec, LayerSpec


@cell
def taper(
    length: float = 10.0,
    width1: float = 0.5,
    width2: float | None = None,
    port: Port | None = None,
    with_bbox: bool = True,
    with_two_ports: bool = True,
    cross_section: CrossSectionSpec = "xs_sc",
    port_order_name: tuple | None = ("o1", "o2"),
    port_order_types: tuple | None = ("optical", "optical"),
    add_pins: bool = True,
    post_process: Callable | None = None,
    **kwargs,
) -> Component:
    """Linear taper, which tapers only the main cross section section.

    Deprecated, use gf.components.taper_cross_section instead

    Args:
        length: taper length.
        width1: width of the west/left port.
        width2: width of the east/right port. Defaults to width1.
        port: can taper from a port instead of defining width1.
        with_bbox: box in bbox_layers and bbox_offsets to avoid DRC sharp edges.
        with_two_ports: includes a second port.
            False for terminator and edge coupler fiber interface.
        cross_section: specification (CrossSection, string, CrossSectionFactory dict).
        port_order_name(tuple): Ordered tuple of port names. First port is default \
                taper port, second name only if with_two_ports flags used.
        port_order_types(tuple): Ordered tuple of port types. First port is default \
                taper port, second name only if with_two_ports flags used.
        add_pins: add pins to the component.
        post_process: function to post process the component.
        kwargs: cross_section settings.
    """
    c = gf.Component()

    width1 = gf.snap.snap_to_grid2x(width1)
    x1 = gf.get_cross_section(cross_section, width=width1)
    if width2:
        width2 = gf.snap.snap_to_grid2x(width2)
        x2 = gf.get_cross_section(cross_section, width=width2)
    else:
        x2 = x1

    width1 = x1.width
    width2 = x2.width

    width_max = max([width1, width2])
    x = gf.get_cross_section(cross_section, width=width_max, **kwargs)
    layer = x.layer

    if isinstance(port, gf.Port) and width1 is None:
        width1 = port.width

    delta_width = width2 - width1

    length = float(snap_to_grid(length))
    y1 = width1 / 2
    y2 = width2 / 2

    if length:
        xpts = [0, length, length, 0]
        ypts = [y1, y2, -y2, -y1]
        c.add_polygon((xpts, ypts), layer=layer)

        xpts = [0, length, length, 0]
        for section in x.sections[1:]:
            layer = section.layer
            if not section.offset:
                y1 = section.width / 2
                y2 = section.width / 2 + delta_width / 2
                ypts = [y1, y2, -y2, -y1]
                c.add_polygon((xpts, ypts), layer=layer)
            else:
                y1 = section.width / 2
                y2 = section.width / 2
                y2 = section.width / 2 + delta_width / 2
                ypts = [y1, y2, -y2, -y1]
                ypts = [y - section.offset for y in ypts]
                c.add_polygon((xpts, ypts), layer=layer)

    c.add_port(
        name=port_order_name[0],
        center=(0, 0),
        width=width1,
        orientation=180,
        layer=x.layer,
        cross_section=x1,
        port_type=port_order_types[0],
    )
    if with_two_ports:
        c.add_port(
            name=port_order_name[1],
            center=(length, 0),
            width=width2,
            orientation=0,
            layer=x.layer,
            cross_section=x2,
            port_type=port_order_types[1],
        )

    if with_bbox and length:
        x.add_bbox(c)
    if add_pins and x.add_pins:
        x.add_pins(c)
    if post_process:
        post_process(c)

    c.info["length"] = float(length)
    c.info["width1"] = float(width1)
    c.info["width2"] = float(width2)
    return c


@gf.cell
def taper_strip_to_ridge(
    length: float = 10.0,
    width1: float = 0.5,
    width2: float = 0.5,
    w_slab1: float = 0.15,
    w_slab2: float = 6.0,
    layer_wg: LayerSpec = "WG",
    layer_slab: LayerSpec = "SLAB90",
    cross_section: CrossSectionSpec = "xs_sc",
    post_process: Callable | None = None,
    **kwargs,
) -> Component:
    r"""Linear taper from strip to rib.

    Deprecated, use gf.components.taper_cross_section instead.

    Args:
        length: taper length (um).
        width1: in um.
        width2: in um.
        w_slab1: slab width in um.
        w_slab2: slab width in um.
        layer_wg: for input waveguide.
        layer_slab: for output waveguide with slab.
        cross_section: for input waveguide.
        post_process: function to post process the component.
        kwargs: cross_section settings.

    .. code::

                      __________________________
                     /           |
             _______/____________|______________
                   /             |
       width1     |w_slab1       | w_slab2  width2
             ______\_____________|______________
                    \            |
                     \__________________________

    """

    width1 = gf.snap.snap_to_grid2x(width1)
    width2 = gf.snap.snap_to_grid2x(width2)
    w_slab1 = gf.snap.snap_to_grid2x(w_slab1)
    w_slab2 = gf.snap.snap_to_grid2x(w_slab2)

    xs = gf.get_cross_section(cross_section, **kwargs)
    xs_wg = gf.get_cross_section(
        cross_section, layer=layer_wg, add_pins_function_name=None
    )
    xs_slab = gf.get_cross_section(
        cross_section, layer=layer_slab, add_pins_function_name=None
    )

    taper_wg = taper(
        length=length,
        width1=width1,
        width2=width2,
        cross_section=xs_wg,
    )
    taper_slab = taper(
        length=length,
        width1=w_slab1,
        width2=w_slab2,
        cross_section=xs_slab,
    )

    c = gf.Component()
    for _t in [taper_wg, taper_slab]:
        taper_ref = _t.ref()
        c.add(taper_ref)
        c.absorb(taper_ref)

    c.info["length"] = float(length)
    c.add_port(name="o1", port=taper_wg.ports["o1"])
    c.add_port(name="o2", port=taper_wg.ports["o2"])

    if xs.add_pins:
        xs.add_pins(c)

    if length:
        xs.add_bbox(c)

    if post_process:
        post_process(c)
    return c


@gf.cell
def taper_strip_to_ridge_trenches(
    length: float = 10.0,
    width: float = 0.5,
    slab_offset: float = 3.0,
    trench_width: float = 2.0,
    trench_layer: LayerSpec = "DEEP_ETCH",
    layer_wg: LayerSpec = "WG",
    trench_offset: float = 0.1,
    post_process: Callable | None = None,
) -> gf.Component:
    """Defines taper using trenches to define the etch.

    Args:
        length: in um.
        width: in um.
        slab_offset: in um.
        trench_width: in um.
        trench_layer: trench layer.
        layer_wg: waveguide layer.
        trench_offset: after waveguide in um.
        post_process: function to post process the component.
    """
    c = gf.Component()
    y0 = width / 2 + trench_width - trench_offset
    yL = width / 2 + trench_width - trench_offset + slab_offset

    # straight
    x = [0, length, length, 0]
    yw = [y0, yL, -yL, -y0]
    if length:
        c.add_polygon((x, yw), layer=layer_wg)

        # top trench
        ymin0 = width / 2
        yminL = width / 2
        ymax0 = width / 2 + trench_width
        ymaxL = width / 2 + trench_width + slab_offset
        x = [0, length, length, 0]
        ytt = [ymin0, yminL, ymaxL, ymax0]
        ytb = [-ymin0, -yminL, -ymaxL, -ymax0]
        c.add_polygon((x, ytt), layer=trench_layer)
        c.add_polygon((x, ytb), layer=trench_layer)

    c.add_port(name="o1", center=(0, 0), width=width, orientation=180, layer=layer_wg)
    c.add_port(
        name="o2", center=(length, 0), width=width, orientation=0, layer=layer_wg
    )
    if post_process:
        post_process(c)
    return c


taper_strip_to_slab150 = partial(taper_strip_to_ridge, layer_slab="SLAB150")

# taper StripCband to NitrideCband
taper_sc_nc = partial(
    taper_strip_to_ridge,
    layer_wg="WG",
    layer_slab="WGN",
    length=20.0,
    width1=0.5,
    width2=0.15,
    w_slab1=0.15,
    w_slab2=1.0,
)


if __name__ == "__main__":
    c = taper(cross_section="xs_rc", width2=1, length=1)
    # xs_pin_m1 = partial(
    #     gf.cross_section.strip_auto_widen,
    #     width=0.5,
    #     width_wide=2,
    #     sections=(
    #         gf.Section(width=1, offset=2, layer=(24, 0), name="n+"),
    #         gf.Section(width=1, offset=3, layer=(41, 0), name="m1"),
    #     ),
    # )
    # # c = taper(width2=10, length=10)
    # # c = taper_strip_to_ridge_trenches()
    # # c = taper_strip_to_ridge()
    # # c = taper(width1=1.5, width2=1, cross_section="xs_rc")
    # # c = taper_sc_nc()
    # c = gf.Component("taper_with_offset")
    # route = gf.routing.get_route_from_waypoints(
    #     [(0, 0), (300, 0), (300, 300), (300, 600), (600, 600)],
    #     # cross_section="xs_sc_auto_widen",
    #     cross_section=xs_pin_m1,
    #     radius=30,
    # )
    # c.add(route.references)
    c.show(show_ports=False)



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.components.grating_coupler_elliptical import grating_coupler_elliptical
from gdsfactory.typings import ComponentSpec


@gf.cell
def grating_coupler_array(
    grating_coupler: ComponentSpec = grating_coupler_elliptical,
    pitch: float = 127.0,
    n: int = 6,
    port_name: str = "o1",
    rotation: int = 0,
    with_loopback: bool = False,
    cross_section: str = "xs_sc",
    bend: ComponentSpec = bend_euler,
    grating_coupler_spacing: float = 0.0,
    **kwargs,
) -> Component:
    """Array of grating couplers.

    Args:
        grating_coupler: ComponentSpec.
        pitch: x spacing.
        n: number of grating couplers.
        port_name: port name.
        rotation: rotation angle for each reference.
        with_loopback: if True, adds a loopback between edge GCs. Only works for rotation = 90 for now.
        cross_section: cross_section for the routing.
        bend: bend component.
        grating_coupler_spacing: spacing between grating couplers.
        kwargs: cross_section settings.
    """
    c = Component()
    grating_coupler = gf.get_component(grating_coupler)

    for i in range(n):
        gc = c << grating_coupler
        gc.rotate(rotation)
        gc.x = i * pitch
        port_name_new = f"o{i+1}"
        c.add_port(port=gc.ports[port_name], name=port_name_new)

    if with_loopback:
        if rotation != 90:
            raise ValueError(
                "with_loopback is currently only programmed to work with rotation = 90"
            )
        routing_xs = gf.get_cross_section(cross_section, **kwargs)
        radius = routing_xs.radius
        bend = bend(radius=radius, angle=180, cross_section=routing_xs)
        sw = gf.c.straight(
            cross_section=routing_xs, length=max(gc.size) + grating_coupler_spacing
        )

        b1 = c << bend
        b2 = c << bend
        b1.mirror()
        b1.connect("o1", c.ports["o1"])
        b2.connect("o1", c.ports[f"o{n}"])

        s1 = c << sw
        s2 = c << sw
        s1.connect("o1", b1.ports["o2"])
        s2.connect("o1", b2.ports["o2"])

        route = gf.routing.get_route(
            s1.ports["o2"],
            s2.ports["o2"],
            cross_section=routing_xs,
        )
        c.add(route.references)
        c.ports.pop("o1")
        c.ports.pop(f"o{n}")

    return c


if __name__ == "__main__":
    # c = grating_coupler_array()
    c = grating_coupler_array(rotation=90, with_loopback=True, radius=50)
    c.show(show_ports=True)



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.rectangle import rectangle
from gdsfactory.typings import ComponentSpec, LayerSpec, LayerSpecs


@cell
def align_wafer(
    width: float = 10.0,
    spacing: float = 10.0,
    cross_length: float = 80.0,
    layer: LayerSpec = "WG",
    layers: LayerSpecs | None = None,
    layer_cladding: tuple[int, int] | None = None,
    square_corner: str = "bottom_left",
) -> Component:
    """Returns cross inside a frame to align wafer.

    Args:
        width: in um.
        spacing: in um.
        cross_length: for the cross.
        layer: for the cross.
        layers: Optional. List of layers for cross.
        layer_cladding: optional.
        square_corner: bottom_left, bottom_right, top_right, top_left.
    """
    c = Component()
    layers = layers or [layer]

    for layer in layers:
        layer = gf.get_layer(layer)
        cross = gf.components.cross(length=cross_length, width=width, layer=layer)
        c.add_ref(cross)

        b = cross_length / 2 + spacing + width / 2
        w = width

        rh = rectangle(size=(2 * b + w, w), layer=layer, centered=True)
        rtop = c.add_ref(rh)
        rbot = c.add_ref(rh)
        rtop.movey(+b)
        rbot.movey(-b)

        rv = rectangle(size=(w, 2 * b), layer=layer, centered=True)
        rl = c.add_ref(rv)
        rr = c.add_ref(rv)
        rl.movex(-b)
        rr.movex(+b)

        wsq = (cross_length + 2 * spacing) / 4
        square_mark = c << rectangle(size=(wsq, wsq), layer=layer, centered=True)
        a = width / 2 + wsq / 2 + spacing

        corner_to_position = {
            "bottom_left": (-a, -a),
            "bottom_right": (a, -a),
            "top_right": (a, a),
            "top_left": (-a, a),
        }

        square_mark.move(corner_to_position[square_corner])

        if layer_cladding:
            rc_tile_excl = rectangle(
                size=(2 * (b + spacing), 2 * (b + spacing)),
                layer=layer_cladding,
                centered=True,
            )
            c.add_ref(rc_tile_excl)

    return c


@cell
def add_frame(
    component: ComponentSpec = rectangle,
    width: float = 10.0,
    spacing: float = 10.0,
    layer: LayerSpec = "WG",
    layers: LayerSpecs | None = None,
) -> Component:
    """Returns component with a frame around it.

    Args:
        component: Component to frame.
        width: of the frame.
        spacing: of component to frame.
        layer: frame layer.
        layers: Optional. List of layers for geometry.
    """
    c = Component()
    component = gf.get_component(component)
    layers = layers or [layer]

    for layer in layers:
        layer = gf.get_layer(layer)
        cref = c.add_ref(component)
        cref.move(-c.size_info.center)
        y = (
            max([component.size_info.height, component.size_info.width]) / 2
            + spacing
            + width / 2
        )
        x = y
        w = width

        rh = rectangle(size=(2 * y + w, w), layer=layer, centered=True)
        rtop = c.add_ref(rh)
        rbot = c.add_ref(rh)
        rtop.movey(+y)
        rbot.movey(-y)

        rv = rectangle(size=(w, 2 * y), layer=layer, centered=True)
        rl = c.add_ref(rv)
        rr = c.add_ref(rv)
        rl.movex(-x)
        rr.movex(+x)
        c.absorb(cref)
    return c


if __name__ == "__main__":
    layers = {(1, 0), (2, 0)}

    # c = gf.components.straight()
    # c = add_frame(component=c)
    c = align_wafer(layers=layers)
    c = add_frame(c, layers=layers)
    c.show(show_ports=True)



"""You can add pins in a pin layer to clearly see the component ports."""

from __future__ import annotations

import gdsfactory as gf
from gdsfactory.typings import LayerSpec


@gf.cell
def straight_narrow(
    length: float = 5.0, width: float = 0.3, layer: LayerSpec = (2, 0)
) -> gf.Component:
    """Returns straight Component.

    Args:
        length: of the straight.
        width: in um.
        layer: layer spec
    """
    wg = gf.Component("straight_sample")
    wg.add_polygon([(0, 0), (length, 0), (length, width), (0, width)], layer=layer)
    wg.add_port(
        name="o1", center=(0, width / 2), width=width, orientation=180, layer=layer
    )
    wg.add_port(
        name="o2", center=(length, width / 2), width=width, orientation=0, layer=layer
    )
    return wg


if __name__ == "__main__":
    from functools import partial

    c = partial(
        gf.c.spiral_inner_io,
        decorator=gf.c.add_grating_couplers_with_loopback_fiber_array,
    )
    wg = c()
    # By default show adds pins, so you don't need it to show_ports
    wg.show(show_ports=False)



from __future__ import annotations

import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.text_rectangular import text_rectangular
from gdsfactory.constants import _glyph, _indent, _width
from gdsfactory.typings import Coordinate, LayerSpec, LayerSpecs


@gf.cell
def text(
    text: str = "abcd",
    size: float = 10.0,
    position: Coordinate = (0, 0),
    justify: str = "left",
    layer: LayerSpec = "WG",
    layers: LayerSpecs | None = None,
) -> Component:
    """Text shapes.

    Args:
        text: string.
        size: in um.
        position: x, y position.
        justify: left, right, center.
        layer: for the text.
        layers: optional for duplicating the text.
    """
    scaling = size / 1000
    xoffset = position[0]
    yoffset = position[1]
    t = gf.Component()
    layers = layers or [layer]

    for line in text.split("\n"):
        label = gf.Component()
        for c in line:
            ascii_val = ord(c)
            if c == " ":
                xoffset += 500 * scaling
            elif 33 <= ascii_val <= 126:
                for poly in _glyph[ascii_val]:
                    xpts = np.array(poly)[:, 0] * scaling
                    ypts = np.array(poly)[:, 1] * scaling
                    for layer in layers:
                        label.add_polygon([xpts + xoffset, ypts + yoffset], layer=layer)
                xoffset += (_width[ascii_val] + _indent[ascii_val]) * scaling
            else:
                raise ValueError(f"No character with ascii value {ascii_val!r}")
        t.add_ref(label)
        yoffset -= 1500 * scaling
        xoffset = position[0]
    justify = justify.lower()
    for label in t.references:
        if justify == "left":
            pass
        elif justify == "right":
            label.xmax = position[0]
        elif justify == "center":
            label.move(origin=label.center, destination=position, axis="x")
        else:
            raise ValueError(
                f"justify = {justify!r} not in ('center', 'right', 'left')"
            )
    return t.flatten()


@gf.cell
def text_lines(
    text: tuple[str, ...] = ("Chip", "01"),
    size: float = 0.4,
    layer: LayerSpec = "WG",
) -> Component:
    """Returns a Component from a text lines.

    Args:
        text: list of strings.
        size: text size.
        layer: text layer.
    """
    c = gf.Component()

    for i, texti in enumerate(text):
        t = text_rectangular(text=texti, size=size, layer=layer)
        tref = c.add_ref(t)
        tref.movey(-6 * size * (i + 1))
    return c


if __name__ == "__main__":
    # c1 = gf.components.text("hello", size=10, layer=(1, 0))
    # c2 = gf.components.text("10.0")
    c = text(
        text=".[,ABCDEFGHIKKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789:/",
        size=4.0,
        justify="right",
        position=(0, 0),
        layers=[(1, 0), (2, 0)],
    )
    # c = text_lines(text=["a", "b"], size=10)
    # c = text_lines()
    # c2.show(show_ports=True)
    # c.plot()
    c.show()



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.straight import straight as straight_function
from gdsfactory.components.taper import taper as taper_function
from gdsfactory.typings import Callable, ComponentFactory, CrossSectionSpec


@gf.cell
def mmi1x2(
    width: float | None = None,
    width_taper: float = 1.0,
    length_taper: float = 10.0,
    length_mmi: float = 5.5,
    width_mmi: float = 2.5,
    gap_mmi: float = 0.25,
    taper: ComponentFactory = taper_function,
    straight: ComponentFactory = straight_function,
    with_bbox: bool = True,
    cross_section: CrossSectionSpec = "xs_sc",
    post_process: Callable | None = None,
    **kwargs,
) -> Component:
    r"""1x2 MultiMode Interferometer (MMI).

    Args:
        width: input and output straight width. Defaults to cross_section width.
        width_taper: interface between input straights and mmi region.
        length_taper: into the mmi region.
        length_mmi: in x direction.
        width_mmi: in y direction.
        gap_mmi:  gap between tapered wg.
        taper: taper function.
        straight: straight function.
        with_bbox: add rectangular box in cross_section
            bbox_layers and bbox_offsets to avoid DRC sharp edges.
        cross_section: specification (CrossSection, string or dict).
        post_process: function to post process the component.
        kwargs: cross_section settings.


    .. code::

               length_mmi
                <------>
                ________
               |        |
               |         \__
               |          __  o2
            __/          /_ _ _ _
         o1 __          | _ _ _ _| gap_mmi
              \          \__
               |          __  o3
               |         /
               |________|

             <->
        length_taper

    """
    c = Component()
    gap_mmi = gf.snap.snap_to_grid(gap_mmi, grid_factor=2)
    x = gf.get_cross_section(cross_section, **kwargs)
    width = width or x.width

    _taper = taper(
        length=length_taper,
        width1=width,
        width2=width_taper,
        cross_section=x,
        add_pins=False,
    )

    delta_width = width_mmi - width
    y = width_mmi / 2
    c.add_polygon([(0, -y), (length_mmi, -y), (length_mmi, y), (0, y)], layer=x.layer)
    for section in x.sections[1:]:
        layer = section.layer
        y = section.width / 2 + delta_width / 2
        c.add_polygon(
            [
                (-delta_width, -y),
                (length_mmi + delta_width, -y),
                (length_mmi + delta_width, y),
                (-delta_width, y),
            ],
            layer=layer,
        )

    a = gap_mmi / 2 + width_taper / 2
    ports = [
        gf.Port(
            "o1",
            orientation=180,
            center=(0, 0),
            width=width_taper,
            layer=x.layer,
            cross_section=x,
        ),
        gf.Port(
            "o2",
            orientation=0,
            center=(+length_mmi, +a),
            width=width_taper,
            layer=x.layer,
            cross_section=x,
        ),
        gf.Port(
            "o3",
            orientation=0,
            center=(+length_mmi, -a),
            width=width_taper,
            layer=x.layer,
            cross_section=x,
        ),
    ]

    for port in ports:
        taper_ref = c << _taper
        taper_ref.connect(port="o2", destination=port)
        c.add_port(name=port.name, port=taper_ref.ports["o1"])
        c.absorb(taper_ref)

    if with_bbox:
        x.add_bbox(c)
    if x.add_pins:
        x.add_pins(c)
    if post_process:
        post_process(c)
    return c


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.components.mmi1x2(cross_section="xs_rc", layer=(2, 0))

    # print(c.xmin)
    # c.xmin = 0
    # print(c.xmin)

    c.show(show_ports=True)



from __future__ import annotations

import numpy as np
from numpy import cos, float64, ndarray, sin

RAD2DEG = 180.0 / np.pi
DEG2RAD = 1 / RAD2DEG


def sign_shape(pts: ndarray) -> float64:
    pts2 = np.roll(pts, 1, axis=0)
    dx = pts2[:, 0] - pts[:, 0]
    y = pts2[:, 1] + pts[:, 1]
    return np.sign((dx * y).sum())


def area(pts: ndarray) -> float64:
    """Returns the area."""
    pts2 = np.roll(pts, 1, axis=0)
    dx = pts2[:, 0] - pts[:, 0]
    y = pts2[:, 1] + pts[:, 1]
    return (dx * y).sum() / 2


def manhattan_direction(p0, p1, tol=1e-5):
    """Returns manhattan direction between 2 points."""
    dp = p1 - p0
    dx, dy = dp[0], dp[1]
    if abs(dx) < tol:
        sx = 0
    elif dx > 0:
        sx = 1
    else:
        sx = -1

    if abs(dy) < tol:
        sy = 0
    elif dy > 0:
        sy = 1
    else:
        sy = -1
    return np.array((sx, sy))


def remove_flat_angles(points: ndarray) -> ndarray:
    a = angles_deg(np.vstack(points))
    da = a - np.roll(a, 1)
    da = np.mod(np.round(da, 3), 180)

    # To make sure we do not remove points at the edges
    da[0] = 1
    da[-1] = 1

    to_rm = list(np.where(np.abs(da[:-1]) < 1e-9)[0])
    if isinstance(points, list):
        while to_rm:
            i = to_rm.pop()
            points.pop(i)

    else:
        points = points[da != 0]

    return points


def remove_identicals(
    pts: ndarray, grids_per_unit: int = 1000, closed: bool = True
) -> ndarray:
    if len(pts) > 1:
        identicals = np.prod(abs(pts - np.roll(pts, -1, 0)) < 0.5 / grids_per_unit, 1)
        if not closed:
            identicals[-1] = False
        pts = np.delete(pts, identicals.nonzero()[0], 0)
    return pts


def centered_diff(a: ndarray) -> ndarray:
    d = (np.roll(a, -1, axis=0) - np.roll(a, 1, axis=0)) / 2
    return d[1:-1]


def centered_diff2(a: ndarray) -> ndarray:
    d = (np.roll(a, -1, axis=0) - a) - (a - np.roll(a, 1, axis=0))
    return d[1:-1]


def curvature(points: ndarray, t: ndarray) -> ndarray:
    """Args are the points and the tangents at each point.

        points : numpy.array shape (n, 2)
        t: numpy.array of size n

    Return:
        The curvature at each point.

    Computes the curvature at every point excluding the first and last point.

    For a planar curve parametrized as P(t) = (x(t), y(t)), the curvature is given
    by (x' y'' - x'' y' ) / (x' **2 + y' **2)**(3/2)

    """
    # Use centered difference for derivative
    dt = centered_diff(t)
    dp = centered_diff(points)
    dp2 = centered_diff2(points)

    dx = dp[:, 0] / dt
    dy = dp[:, 1] / dt

    dx2 = dp2[:, 0] / dt**2
    dy2 = dp2[:, 1] / dt**2

    return (dx * dy2 - dx2 * dy) / (dx**2 + dy**2) ** (3 / 2)


def radius_of_curvature(points, t):
    return 1 / curvature(points, t)


def path_length(points: ndarray) -> float64:
    """Returns: The path length.

    Args:
        points: With shape (N, 2) representing N points with coordinates x, y.
    """
    dpts = points[1:, :] - points[:-1, :]
    _d = dpts**2
    return np.sum(np.sqrt(_d[:, 0] + _d[:, 1]))


def snap_angle(a: float64) -> int:
    """Returns angle snapped along manhattan angle (0, 90, 180, 270).

    a: angle in deg
    Return angle snapped along manhattan angle
    """
    a = a % 360
    if -45 < a < 45:
        return 0
    elif 45 < a < 135:
        return 90
    elif 135 < a < 225:
        return 180
    elif 225 < a < 315:
        return 270
    else:
        return 0


def angles_rad(pts: ndarray) -> ndarray:
    """Returns the angles (radians) of the connection between each point and the next."""
    _pts = np.roll(pts, -1, 0)
    return np.arctan2(_pts[:, 1] - pts[:, 1], _pts[:, 0] - pts[:, 0])


def angles_deg(pts: ndarray) -> ndarray:
    """Returns the angles (degrees) of the connection between each point and the next."""
    return angles_rad(pts) * RAD2DEG


def extrude_path(
    points: ndarray,
    width: float,
    with_manhattan_facing_angles: bool = True,
    spike_length: float64 | int | float = 0,
    start_angle: int | None = None,
    end_angle: int | None = None,
    grid: float | None = None,
) -> ndarray:
    """Deprecated. Use gdsfactory.path.Path.extrude() instead.

    Extrude a path of `width` along a curve defined by `points`.

    Args:
        points: numpy 2D array of shape (N, 2).
        width: of the path to extrude.
        with_manhattan_facing_angles: snaps to manhattan angles.
        spike_length: in um.
        start_angle: in degrees.
        end_angle: in degrees.
        grid: in um.

    Returns:
        numpy 2D array of shape (2*N, 2).
    """
    from gdsfactory.pdk import get_grid_size

    grid = grid or get_grid_size()

    if isinstance(points, list):
        points = np.stack([(p[0], p[1]) for p in points], axis=0)

    a = angles_deg(points)
    if with_manhattan_facing_angles:
        _start_angle = snap_angle(a[0] + 180)
        _end_angle = snap_angle(a[-2])
    else:
        _start_angle = a[0] + 180
        _end_angle = a[-2]

    start_angle = start_angle if start_angle is not None else _start_angle
    end_angle = end_angle if end_angle is not None else _end_angle

    a2 = angles_rad(points) * 0.5
    a1 = np.roll(a2, 1)

    a2[-1] = end_angle * DEG2RAD - a2[-2]
    a1[0] = start_angle * DEG2RAD - a1[1]

    a_plus = a2 + a1
    cos_a_min = np.cos(a2 - a1)
    offsets = np.column_stack((-sin(a_plus) / cos_a_min, cos(a_plus) / cos_a_min)) * (
        0.5 * width
    )

    points_back = np.flipud(points - offsets)
    if spike_length != 0:
        d = spike_length
        a_start = start_angle * DEG2RAD
        a_end = end_angle * DEG2RAD
        p_start_spike = points[0] + d * np.array([[cos(a_start), sin(a_start)]])
        p_end_spike = points[-1] + d * np.array([[cos(a_end), sin(a_end)]])

        pts = np.vstack((p_start_spike, points + offsets, p_end_spike, points_back))
    else:
        pts = np.vstack((points + offsets, points_back))

    pts = np.round(pts / grid) * grid

    return pts


def polygon_grow(polygon: ndarray, offset: float) -> ndarray:
    """Returns a grown closed shaped polygon by an offset."""
    s = remove_identicals(polygon)
    s = remove_flat_angles(s)
    s = np.vstack([s, s[0]])
    if len(s) <= 1:
        return s

    # Make sure the shape is oriented in the correct direction for scaling
    ss = sign_shape(s)
    offset *= -ss

    a2 = angles_rad(s) * 0.5
    a1 = np.roll(a2, 1)

    a2[-1] = a2[0]
    a1[0] = a1[-1]

    a = a2 + a1
    c_minus = cos(a2 - a1)
    offsets = np.column_stack((-sin(a) / c_minus, cos(a) / c_minus)) * offset
    return s + offsets



from __future__ import annotations

from functools import partial

import gdsfactory as gf
from gdsfactory.components.via_stack import (
    via_stack,
    via_stack_corner45_extended,
    via_stack_m1_m3,
)
from gdsfactory.snap import snap_to_grid

Float2 = tuple[float, float]
Coordinate = tuple[Float2, Float2]


@gf.cell
def seal_ring(
    bbox=((-1.0, -1.0), (3.0, 4.0)),
    seal: gf.typings.ComponentSpec = via_stack,
    width: float = 10,
    padding: float = 10.0,
    with_north: bool = True,
    with_south: bool = True,
    with_east: bool = True,
    with_west: bool = True,
) -> gf.Component:
    """Returns a continuous seal ring boundary at the chip/die.

    Prevents cracks from spreading and shields when connected to ground.

    Args:
        bbox: to add seal ring around. You can pass Component.bbox.
        seal: function for the seal.
        width: of the seal.
        padding: from component to seal.
        with_north: includes seal.
        with_south: includes seal.
        with_east: includes seal.
        with_west: includes seal.
    """
    c = gf.Component()
    (xmin, ymin), (xmax, ymax) = bbox

    x = (xmax + xmin) / 2

    sx = xmax - xmin
    sy = ymax - ymin

    snap = partial(snap_to_grid, grid_factor=2)
    sx = snap(sx)
    sy = snap(sy)

    ymin_north = snap(ymax + padding)
    ymax_south = snap(ymax - sy - padding)

    # north south
    size_north_south = (sx + 2 * padding + 2 * width, width)
    size_east_west = (width, sy + 2 * padding)

    if with_north:
        north = c << seal(size=size_north_south)
        north.ymin = ymin_north
        north.x = x

    if with_east:
        east = c << seal(size=size_east_west)
        east.xmin = xmax + padding
        east.ymax = ymin_north

    if with_west:
        west = c << seal(size=size_east_west)
        west.xmax = xmin - padding
        west.ymax = ymin_north

    if with_south:
        south = c << seal(size=size_north_south)
        south.ymax = ymax_south
        south.x = x

    return c


@gf.cell
def seal_ring_segmented(
    bbox=((-1.0, -1.0), (3.0, 4.0)),
    length_segment: float = 10,
    width_segment: float = 3,
    spacing_segment: float = 2,
    corner: gf.Component = via_stack_corner45_extended,
    via_stack: gf.Component = via_stack_m1_m3,
    with_north: bool = True,
    with_south: bool = True,
    with_east: bool = True,
    with_west: bool = True,
    padding: float = 10.0,
) -> gf.Component:
    """Segmented Seal ring.

    Args:
        size: size of the seal ring.
        length_segment: length of each segment.
        width_segment: width of each segment.
        spacing_segment: spacing between segments.
        corner: corner component.
        via_stack: via_stack component.
        with_north: includes seal.
        with_south: includes seal.
        with_east: includes seal.
        with_west: includes seal.
        padding: from component to seal.
    """
    c = gf.Component()
    corner = gf.get_component(corner, width=width_segment)

    (xmin, ymin), (xmax, ymax) = bbox

    xmin -= padding
    xmax += padding
    ymin -= padding
    ymax += padding

    tl = c << corner
    tr = c << corner

    tl.xmin = xmin
    tl.ymax = ymax

    tr.mirror()
    tr.xmax = xmax
    tr.ymax = ymax

    bl = c << corner
    br = c << corner
    br.mirror()
    br.mirror_y()
    bl.mirror_y()

    bl.xmin = xmin
    bl.ymin = ymin

    br.xmax = xmax
    br.ymin = ymin

    pitch = length_segment + spacing_segment

    # horizontal
    dx = abs(tl.xmax - tr.xmin)
    segment_horizontal = via_stack(size=(length_segment, width_segment))

    horizontal = gf.c.array(
        component=segment_horizontal,
        columns=int(dx / pitch),
        spacing=(pitch, 0),
    )

    if with_north:
        top = c << horizontal
        top.ymax = tl.ymax
        top.xmin = tl.xmax + spacing_segment

        # horizontal inner
        topi = c << horizontal
        topi.ymax = top.ymin - spacing_segment
        topi.xmin = top.xmin + pitch / 2

    if with_south:
        bot = c << horizontal
        bot.ymin = ymin
        bot.xmin = tl.xmax + spacing_segment

        boti = c << horizontal
        boti.ymin = bot.ymax + spacing_segment
        boti.xmin = bot.xmin + spacing_segment

    # vertical
    segment_vertical = via_stack(size=(width_segment, length_segment))
    dy = abs(tl.ymin - bl.ymax)

    vertical = gf.c.array(
        component=segment_vertical,
        rows=int(dy / pitch),
        columns=1,
        spacing=(0, pitch),
    )

    if with_east:
        right = c << vertical
        right.xmax = xmax
        right.ymin = bl.ymax
        righti = c << vertical
        righti.xmax = right.xmin - spacing_segment
        righti.ymin = right.ymin + pitch / 2

    if with_west:
        left = c << vertical
        left.xmin = xmin
        left.ymin = bl.ymax

        # vertical inner
        lefti = c << vertical
        lefti.xmin = left.xmax + spacing_segment
        lefti.ymin = left.ymin + pitch / 2

    return c


if __name__ == "__main__":
    c = gf.Component()
    ref = c << gf.c.rectangle(size=(500, 100), layer=(1, 0))
    ref.move((500, 300))
    c << seal_ring_segmented(c.bbox, with_south=False, padding=50, spacing_segment=5)
    # big_square = partial(rectangle, size=(1300, 2600))
    # c = gf.Component("demo")
    # c << big_square()
    # c << seal_ring(c.bbox + ((0, 0), (10, 0)), with_south=False)
    c.show(show_ports=True)



from functools import partial

import gdstk

import gdsfactory as gf
from gdsfactory.typings import (
    Component,
    ComponentOrReference,
    ComponentSpec,
    Floats,
    LayerSpec,
    LayerSpecs,
)


def get_polygons_over_under(
    component: ComponentOrReference,
    layers: LayerSpecs,
    distances: Floats,
    use_union: bool = True,
    precision: float = 1e-4,
    join: str = "miter",
    tolerance: int = 2,
) -> list[gdstk.Polygon]:
    """Returns list polygons dilated and eroded by an offset.
    Cleans min width gap and acute angle DRC errors equal to distances.

    Args:
        component: Component containing polygons to offset.
        layers: list of layers to remove min gap errors.
        distances: Distance to offset polygons.
            By expanding and shrinking it merges polygons.
            If you want to fix min gaps of 0.5 use 0.5.
        precision: Desired precision for rounding vertex coordinates.
        join: {'miter', 'bevel', 'round'} Type of join used to create polygon offset
        tolerance: For miter joints, this number must be at least 2 represents the
          maximal distance in multiples of offset between new vertices and their
          original position before beveling to avoid spikes at acute joints. For
          round joints, it indicates the curvature resolution in number of
          points per full circle.
        layer: Specific layer to put polygon geometry on.
    Returns:
        Component containing a polygon(s) with the specified offset applied.
    """
    polygons = []

    for layer, distance in zip(layers, distances):
        layer = gf.get_layer(layer)
        if layer in component.layers:
            gds_layer, gds_datatype = layer
            polygons_to_offset = component.get_polygons(layer)
            p_dilated = gdstk.offset(
                polygons_to_offset,
                distance=+distance,
                join=join,
                tolerance=tolerance,
                precision=precision,
                use_union=use_union,
                layer=gds_layer,
                datatype=gds_datatype,
            )
            p = gdstk.offset(
                p_dilated,
                distance=-distance,
                join=join,
                tolerance=tolerance,
                precision=precision,
                use_union=use_union,
                layer=gds_layer,
                datatype=gds_datatype,
            )
            polygons.append(p)
    return polygons


@gf.cell_with_child
def over_under(
    component: ComponentSpec,
    layers: LayerSpecs,
    remove_original: bool = False,
    **kwargs,
) -> Component:
    """Returns cleaned component.

    Args:
        component:
        layers: list of layers.
        remove_original: remove original layers.
    """
    c = Component()
    _component = gf.get_component(component)

    ref = c << _component
    if remove_original:
        c = c.remove_layers(layers=layers)
    p = get_polygons_over_under(component=_component, layers=layers, **kwargs)
    c.add(p)
    c.copy_child_info(_component)
    c.add_ports(ref.ports)
    return c


over_under_remove_original = partial(over_under, remove_original=True)


def get_polygons_underplot(
    component: ComponentOrReference,
    layers_extended: LayerSpecs,
    layer_reference: LayerSpec,
    distance: Floats,
    keep_overplot: bool = True,
    use_union: bool = True,
    precision: float = 1e-4,
    join: str = "miter",
    tolerance: int = 2,
) -> list[gdstk.Polygon]:
    """Returns polygons on "layers_extended" .
    Cleans min width gap and acute angle DRC errors equal to distances.

    Args:
        component: Component containing polygons to offset.
        layers: list of layers to remove min gap errors.
        distances: Distance to offset polygons.
            By expanding and shrinking it merges polygons.
            If you want to fix min gaps of 0.5 use 0.5.
        precision: Desired precision for rounding vertex coordinates.
        join: {'miter', 'bevel', 'round'} Type of join used to create polygon offset
        tolerance: For miter joints, this number must be at least 2 represents the
          maximal distance in multiples of offset between new vertices and their
          original position before beveling to avoid spikes at acute joints. For
          round joints, it indicates the curvature resolution in number of
          points per full circle.
        layer: Specific layer to put polygon geometry on.
    Returns:
        Component containing a polygon(s) with the specified offset applied.
    """
    polygons = []

    # Acquire and offset the reference polygons
    layer_reference = gf.get_layer(layer_reference)
    if layer_reference in component.layers:
        gds_layer, gds_datatype = layer_reference
        reference_polygons = component.get_polygons(layer_reference)
        reference_polygons_shrunk = gdstk.offset(
            reference_polygons,
            distance=-distance,
            join=join,
            tolerance=tolerance,
            precision=precision,
            use_union=use_union,
            layer=gds_layer,
            datatype=gds_datatype,
        )

    # For each other layer, trim the polygons
    for layer in layers_extended:
        layer = gf.get_layer(layer)
        if layer in component.layers:
            gds_layer, gds_datatype = layer
            polygons_to_trim = component.get_polygons(layer)
            p = gdstk.boolean(
                polygons_to_trim,
                reference_polygons_shrunk,
                operation="and",
                precision=precision,
                layer=gds_layer,
                datatype=gds_datatype,
            )
            polygons.append(p)
            # Optionally add portion of the polygons outside the reference
            if keep_overplot:
                p = gdstk.boolean(
                    polygons_to_trim,
                    reference_polygons,
                    operation="not",
                    precision=precision,
                    layer=gds_layer,
                    datatype=gds_datatype,
                )
                polygons.append(p)
    return polygons


@gf.cell_with_child
def fix_underplot(
    component: ComponentSpec,
    layers_extended: LayerSpecs,
    layer_reference: LayerSpecs,
    distance: Floats,
    keep_overplot: bool = True,
    **kwargs,
) -> Component:
    """Returns component with polygon underplots.

    Args:
        component:
        layers_extended: list of layers to change.
        layer_reference: layers to underplot w.r.t. (union of these layers is considered)
        distances: list of floats, amount to overplot by
        remove_original: remove original layers.
    """
    c = Component()
    _component = gf.get_component(component)

    ref = c << _component
    c = c.remove_layers(layers=layers_extended)
    p = get_polygons_underplot(
        component=_component,
        layers_extended=layers_extended,
        layer_reference=layer_reference,
        distance=distance,
        keep_overplot=keep_overplot,
        **kwargs,
    )
    c.add(p)
    c.copy_child_info(_component)
    c.add_ports(ref.ports)
    return c


if __name__ == "__main__":
    import gdsfactory as gf

    # over_under_slab = partial(over_under, layers=((2, 0)), distances=(0.5,))
    # c = gf.components.coupler_ring(
    #     cladding_layers=((2, 0)),
    #     cladding_offsets=(0.2,),
    #     decorator=over_under_slab,
    # )
    # over_under_slab_remove_original = partial(
    #     over_under_remove_original, layers=((2, 0)), distances=(0.5,)
    # )
    # c = gf.components.coupler_ring(
    #     cladding_layers=((2, 0)),
    #     cladding_offsets=(0.2,),
    #     decorator=over_under_slab_remove_original,
    # )
    # get_polygons_over_under_slab = partial(
    #     get_polygons_over_under, layers=((2, 0)), distances=(0.5,)
    # )

    # c = gf.Component("component_clean")
    # ref = c << gf.components.coupler_ring(
    #     cladding_layers=((2, 0)),
    #     cladding_offsets=(0.2,),  # decorator=over_under_slab_decorator
    # )
    # polygons = get_polygons_over_under_slab(ref)
    # c.add(polygons)

    # Make component requiring overplot
    c1 = gf.Component("component_initial")
    c1 << gf.components.rectangle(size=(4, 4), layer="WG")
    c1 << gf.components.rectangle(size=(2, 2), layer="SLAB150")
    slab = c1 << gf.components.rectangle(size=(2, 2), layer="SLAB90")
    slab.move((3, 1))
    c1.show(show_ports=True)

    c2 = gf.Component("component_clean")
    c2 = fix_underplot(
        component=c1,
        layers_extended=("SLAB150", "SLAB90"),
        layer_reference="WG",
        distance=0.1,
        keep_overplot=False,
    )
    c2.show(show_ports=True)



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.straight import straight
from gdsfactory.typings import CrossSectionSpec


@gf.cell
def straight_array(
    n: int = 4,
    spacing: float = 4.0,
    length: float = 10.0,
    cross_section: CrossSectionSpec = "xs_sc",
) -> Component:
    """Array of straights connected with grating couplers.

    useful to align the 4 corners of the chip

    Args:
        n: number of straights.
        spacing: edge to edge straight spacing.
        length: straight length (um).
        cross_section: specification (CrossSection, string or dict).
    """
    c = Component()
    wg = straight(cross_section=cross_section, length=length)

    for i in range(n):
        wref = c.add_ref(wg)
        wref.y += i * (spacing + wg.info["width"])
        c.add_ports(wref.ports, prefix=str(i))

    c.auto_rename_ports()
    return c


if __name__ == "__main__":
    c = straight_array()
    # c.pprint_ports()
    c.show(show_ports=True)



from __future__ import annotations

import warnings

import gdsfactory as gf
from gdsfactory.components.straight import straight
from gdsfactory.difftest import difftest
from gdsfactory.port import Port
from gdsfactory.typings import ComponentSpec, Routes


def get_routes_straight(
    ports: list[Port] | dict[str, Port],
    straight: ComponentSpec = straight,
    **kwargs,
) -> Routes:
    """Returns routes made by 180 degree straights.

    Args:
        ports: List or dict of ports.
        straight: function for straight.
        kwargs: waveguide settings.

    .. plot::
        :include-source:

        import gdsfactory as gf

        c = gf.Component("get_routes_straight")
        pad_array = gf.components.pad_array()
        c1 = c << pad_array
        c2 = c << pad_array
        c2.ymax = -200

        routes = gf.routing.get_routes_straight(ports=c1.get_ports_list(), length=200)
        c.add(routes.references)
        c.plot()

    """
    warnings.warn(
        "This function is deprecated and will be removed in next major release."
    )
    ports = list(ports.values()) if isinstance(ports, dict) else ports
    straight = straight(**kwargs)
    references = [straight.ref() for _ in ports]
    references = [
        ref.connect("o1", port, allow_width_mismatch=True)
        for port, ref in zip(ports, references)
    ]
    ports = [ref.ports["o2"] for ref in references]
    lengths = [straight.info["length"]] * len(ports)
    return Routes(references=references, ports=ports, lengths=lengths)


def test_get_routes_straight(check: bool = True) -> None:
    c = gf.Component("get_routes_straight")
    pad_array = gf.components.pad_array()
    c1 = c << pad_array
    c2 = c << pad_array
    c2.ymax = -200

    routes = get_routes_straight(ports=c1.get_ports_list(), length=200)
    c.add(routes.references)
    if check:
        difftest(c)


if __name__ == "__main__":
    c = gf.Component("get_routes_straight")
    pad_array = gf.components.pad_array()
    c1 = c << pad_array
    c2 = c << pad_array
    c2.ymax = -200

    routes = get_routes_straight(ports=c1.get_ports_list(), length=200)
    c.add(routes.references)
    c.show()



from __future__ import annotations

from collections.abc import Iterable
from functools import partial

import numpy as np

import gdsfactory as gf
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.components.straight import straight
from gdsfactory.components.taper import taper
from gdsfactory.components.via_corner import via_corner
from gdsfactory.port import Port
from gdsfactory.routing.get_route_sbend import get_route_sbend
from gdsfactory.routing.manhattan import RouteError, get_route_error, round_corners
from gdsfactory.typings import (
    STEP_DIRECTIVES,
    ComponentSpec,
    CrossSectionSpec,
    MultiCrossSectionAngleSpec,
    Route,
)


def get_route_from_steps(
    port1: Port,
    port2: Port,
    steps: Iterable[dict[str, float]] | None = None,
    bend: ComponentSpec = bend_euler,
    straight: ComponentSpec = straight,
    taper: ComponentSpec | None = taper,
    cross_section: CrossSectionSpec | MultiCrossSectionAngleSpec = "xs_sc",
    with_sbend: bool = True,
    **kwargs,
) -> Route:
    """Returns a route formed by the given waypoints steps.

    Uses smooth euler bends instead of corners and tapers in straight sections.
    Tapering to wider straights reduces the optical loss when auto_widen=True.
    `get_route_from_steps` is a manual version of `get_route`
    and a more concise and convenient version of `get_route_from_waypoints`

    Args:
        port1: start port.
        port2: end port.
        steps: that define the route (x, y, dx, dy) [{'dx': 5}, {'dy': 10}].
        bend: function that returns bends.
        straight: straight spec.
        taper: taper spec.
        cross_section: cross_section spec.
        with_sbend: whether to add sbend for impossible manhattan routes.
        kwargs: cross_section settings.

    .. plot::
        :include-source:

        import gdsfactory as gf

        c = gf.Component("get_route_from_steps_sample")
        w = gf.components.straight()
        left = c << w
        right = c << w
        right.move((100, 80))

        obstacle = gf.components.rectangle(size=(100, 10), port_type=None)
        obstacle1 = c << obstacle
        obstacle2 = c << obstacle
        obstacle1.ymin = 40
        obstacle2.xmin = 25

        p1 = left.ports['o2']
        p2 = right.ports['o2']
        route = gf.routing.get_route_from_steps(
            port1=p1,
            port2=p2,
            steps=[
                {"x": 20},
                {"y": 20},
                {"x": 120},
                {"y": 80},
            ],
        )
        c.add(route.references)
        c.plot()

    """
    x, y = port1.center
    x2, y2 = port2.center

    waypoints = [(x, y)]
    steps = steps or []

    for d in steps:
        if not STEP_DIRECTIVES.issuperset(d):
            invalid_step_directives = list(set(d.keys()) - STEP_DIRECTIVES)
            raise ValueError(
                f"Invalid step directives: {invalid_step_directives}."
                f"Valid directives are {list(STEP_DIRECTIVES)}"
            )
        x = d["x"] if "x" in d else x
        x += d.get("dx", 0)
        y = d["y"] if "y" in d else y
        y += d.get("dy", 0)
        waypoints += [(x, y)]

    waypoints += [(x2, y2)]
    waypoints = np.array(waypoints)

    if isinstance(cross_section, list | tuple):
        xs_list = []
        for element in cross_section:
            xs, angles = element
            xs = gf.get_cross_section(xs)
            xs = gf.get_cross_section(xs, **kwargs)
            xs_list.append((xs, angles))
        cross_section = xs_list

    else:
        xs = cross_section = gf.get_cross_section(cross_section, **kwargs)
        auto_widen = cross_section.auto_widen

        if auto_widen:
            taper = gf.get_component(
                taper,
                length=xs.taper_length,
                width1=xs.width,
                width2=xs.width_wide,
                cross_section=cross_section,
                **kwargs,
            )
        else:
            taper = None

    try:
        return round_corners(
            points=waypoints,
            bend=bend,
            straight=straight,
            taper=taper,
            cross_section=cross_section,
            with_sbend=with_sbend,
            **kwargs,
        )
    except RouteError:
        if with_sbend:
            return get_route_sbend(port1, port2, cross_section=cross_section, **kwargs)
    return get_route_error(points=waypoints, with_sbend=False)


get_route_from_steps_electrical = partial(
    get_route_from_steps, bend="wire_corner", taper=None, cross_section="xs_m3"
)

get_route_from_steps_electrical_multilayer = partial(
    get_route_from_steps,
    bend=via_corner,
    taper=None,
    cross_section=[
        (gf.cross_section.metal2, (90, 270)),
        ("xs_metal_routing", (0, 180)),
    ],
)


def test_route_from_steps():
    c = gf.Component()
    layer = (2, 0)
    w = gf.components.straight(layer=layer)
    left = c << w
    right = c << w
    right.move((100, 80))

    obstacle = gf.components.rectangle(size=(100, 10))
    obstacle1 = c << obstacle
    obstacle2 = c << obstacle
    obstacle1.ymin = 40
    obstacle2.xmin = 25

    p1 = left.ports["o2"]
    p2 = right.ports["o2"]

    route = get_route_from_steps(
        port1=p1,
        port2=p2,
        steps=[
            {"x": 20, "y": 0},
            {"x": 20, "y": 20},
            {"x": 120, "y": 20},
            {"x": 120, "y": 80},
        ],
        layer=layer,
    )

    length = 186.548

    assert route.length == length, route.length
    route = gf.routing.get_route_from_steps(
        port1=p1,
        port2=p2,
        steps=[
            {"x": 20},
            {"y": 20},
            {"x": 120},
            {"y": 80},
        ],
        layer=layer,
    )
    c.add(route.references)
    assert route.length == length, route.length
    return c


if __name__ == "__main__":
    test_route_from_steps()

    # c = gf.Component("get_route_from_steps_sample")
    # w = gf.components.straight()
    # left = c << w
    # right = c << w
    # right.move((100, 80))

    # p1 = left.ports["o2"]
    # p2 = right.ports["o2"]

    # route = get_route_from_steps(
    #     port1=p2,
    #     port2=p1,
    #     steps=[
    #         {"x": 20, "y": 0},
    #         {"x": 20, "y": 20},
    #         {"x": 120, "y": 20},
    #         {"x": 120, "y": 80},
    #     ],
    # )
    # c.add(route.references)
    # c.add(route.labels)
    # c.show(show_ports=True)

    c = gf.Component("pads_route_from_steps")
    pt = c << gf.components.pad_array(orientation=270, columns=3)
    pb = c << gf.components.pad_array(orientation=90, columns=3)
    pt.move((100, 200))
    route = gf.routing.get_route_from_steps_electrical(
        pb.ports["e11"],
        pt.ports["e11"],
        steps=[
            {"y": 200},
            # {"z": 200},
        ],
        # cross_section='metal_routing',
        # bend=gf.components.wire_corner,
    )
    c.add(route.references)
    c.show(show_ports=True)



"""High speed GSG pads."""

from __future__ import annotations

from functools import partial

import gdsfactory as gf
from gdsfactory.components.pad import pad as pad_function
from gdsfactory.components.rectangle import rectangle
from gdsfactory.typings import ComponentSpec, Float2, LayerSpec

rectangle_m3 = partial(rectangle, layer="MTOP")


@gf.cell
def pad_gsg_short(
    via_stack: ComponentSpec = rectangle_m3,
    size: Float2 = (22, 7),
    layer_metal: LayerSpec = "MTOP",
    metal_spacing: float = 5.0,
    short: bool = True,
    pad: ComponentSpec = pad_function,
    pad_spacing: float = 150,
) -> gf.Component:
    """Returns high speed GSG pads for calibrating the RF probes.

    Args:
        via_stack: where the RF pads connect to.
        size: for the via_stack.
        layer_metal: for the short.
        metal_spacing: in um.
        short: if False returns an open.
        pad: function for pad.
        pad_spacing: in um.
    """
    c = gf.Component()

    via = gf.get_component(via_stack, size=size)
    gnd_top = c << via

    if short:
        short = c << rectangle(size=size, layer=layer_metal)
    gnd_bot = c << via

    gnd_bot.ymax = via.ymin
    gnd_top.ymin = via.ymax

    gnd_top.movex(-metal_spacing)
    gnd_bot.movex(-metal_spacing)

    pads = c << gf.components.array(pad, columns=1, rows=3, spacing=(0, pad_spacing))
    pads.xmin = via.xmax + 50
    pads.y = 0

    c << gf.routing.route_quad(
        gnd_bot.ports["e4"], pads.ports["e1_1_1"], layer=layer_metal
    )
    c << gf.routing.route_quad(
        gnd_top.ports["e2"], pads.ports["e1_3_1"], layer=layer_metal
    )
    c << gf.routing.route_quad(via.ports["e3"], pads.ports["e1_2_1"], layer=layer_metal)
    return c


pad_gsg_open = partial(pad_gsg_short, short=False)


if __name__ == "__main__":
    # c = pad_array_double()
    c = pad_gsg_short()
    # c = pad_gsg_open()
    c.show(show_ports=True)



import gdsfactory as gf
from gdsfactory.components.grating_coupler_elliptical import (
    grating_coupler_elliptical_te,
)
from gdsfactory.components.mzi import mzi_phase_shifter
from gdsfactory.components.pad import pad_small
from gdsfactory.typings import Callable, ComponentSpec, CrossSectionSpec


@gf.cell_with_child
def add_fiber_array_optical_south_electrical_north(
    component: ComponentSpec = mzi_phase_shifter,
    pad: ComponentSpec = pad_small,
    with_loopback: bool = True,
    pad_spacing: float = 100.0,
    fiber_spacing: float = 127.0,
    pad_gc_spacing: float = 250.0,
    electrical_port_names: list[str] | None = None,
    electrical_port_orientation: float | None = 90,
    npads: int | None = None,
    pad_assigments: tuple[tuple[str, str], ...] | None = None,
    grating_coupler: ComponentSpec = grating_coupler_elliptical_te,
    xs_metal: CrossSectionSpec = "xs_metal_routing",
    post_process: Callable | None = None,
    **kwargs,
) -> gf.Component:
    """Returns a fiber array with Optical gratings on South and Electrical pads on North.

    This a test configuration for DC pads.

    Args:
        component: component spec to add fiber and pads.
        pad: pad spec.
        with_loopback: whether to add a loopback port.
        pad_spacing: spacing between pads.
        fiber_spacing: spacing between grating couplers.
        pad_gc_spacing: spacing between pads and grating couplers.
        electrical_port_names: list of electrical port names. Defaults to all.
        electrical_port_orientation: orientation of electrical ports. Defaults to 90.
        npads: number of pads. Defaults to one per electrical_port_names.
        pad_assigments: if not None, routes according to (component_port_name, pad_port_name).
        grating_coupler: grating coupler function.
        xs_metal: metal cross section.
        post_process: function to run after the cutback is created.

    Keyword Args:
        gc_port_name: grating coupler input port name.
        gc_port_labels: grating coupler list of labels.
        io_rotation: fiber coupler rotation in degrees. Defaults to None.
        component_name: optional for the label.
        select_ports: function to select ports.
        cross_section: cross_section function.
        get_input_labels_function: function to get input labels. None skips labels.
        layer_label: optional layer for grating coupler label.
        bend: bend spec.
        straight: straight spec.
        taper: taper spec.
        get_input_label_text_loopback_function: function to get input label test.
        get_input_label_text_function: for labels.
        fanout_length: if None, automatic calculation of fanout length.
        max_y0_optical: in um.
        with_loopback: True, adds loopback structures.
        straight_separation: from edge to edge.
        list_port_labels: None, adds TM labels to port indices in this list.
        connected_port_list_ids: names of ports only for type 0 optical routing.
        nb_optical_ports_lines: number of grating coupler lines.
        force_manhattan: False
        excluded_ports: list of port names to exclude when adding gratings.
        grating_indices: list of grating coupler indices.
        routing_straight: function to route.
        routing_method: get_route.
        optical_routing_type: None: auto, 0: no extension, 1: standard, 2: check.
        gc_rotation: fiber coupler rotation in degrees. Defaults to -90.
        input_port_indexes: to connect.

    """
    c = gf.Component()
    component = gf.get_component(component)
    r = c << gf.routing.add_fiber_array(
        component=component,
        grating_coupler=grating_coupler,
        with_loopback=with_loopback,
        fiber_spacing=fiber_spacing,
        layer_label=None,
        **kwargs,
    )
    optical_ports = r.get_ports_list(port_type="optical")
    c.add_ports(optical_ports)

    electrical_ports = r.get_ports_list(
        port_type="electrical", orientation=electrical_port_orientation
    )

    if not electrical_ports:
        raise ValueError(
            f"No electrical ports found with orientation {electrical_port_orientation}. "
            f"{r.pprint_ports()}"
        )

    electrical_port_names = electrical_port_names or [p.name for p in electrical_ports]

    npads = npads or len(electrical_port_names)
    pads = c << gf.components.array(
        component=pad,
        columns=npads,
        spacing=(pad_spacing, 0),
    )
    pads.x = r.x
    pads.ymin = r.ymin + pad_gc_spacing

    electrical_ports = [r[por_name] for por_name in electrical_port_names]
    nroutes = min(len(electrical_ports), npads)

    if electrical_port_orientation and int(electrical_port_orientation) != 90:
        routes, electrical_ports = gf.routing.route_ports_to_side(
            ports=electrical_ports,
            side="north",
            cross_section=xs_metal,
            separation=pad_spacing,
        )
        for route in routes:
            c.add(route.references)

    ports1 = electrical_ports[:nroutes]
    pad_ports = pads.get_ports_list(orientation=270)

    if pad_assigments is None:
        ports2 = pads.get_ports_list(orientation=270)[:nroutes]
        routes = gf.routing.get_bundle_electrical(
            ports1=ports1,
            ports2=ports2,
            cross_section=xs_metal,
            enforce_port_ordering=False,
        )
    else:
        ports2 = []
        ports2_dict = pads.get_ports_dict()
        routes = []
        for port1_name, port2_name in pad_assigments:
            port1 = r[port1_name]
            port2 = ports2_dict[port2_name]
            route = gf.routing.get_route_electrical(input_port=port1, output_port=port2)
            routes.append(route)
            ports2.append(port2)

    for route in routes:
        c.add(route.references)

    c.add_ports(pad_ports)
    c.copy_child_info(component)

    if post_process:
        post_process(c)
    return c


if __name__ == "__main__":
    # c = add_fiber_array_optical_south_electrical_north(
    #     info=dict(wavelength_min=1550), name="my_mzi"
    # )

    # d = json.loads(c.labels[0].text)
    # print(d)
    # import gdsfactory as gf
    # from functools import partial

    # component = partial(mzi_phase_shifter, length_y=1)
    # c = add_fiber_array_optical_south_electrical_north(
    #     component=component,
    #     electrical_port_names=["top_l_e2", "top_r_e2"],
    #     npads=5,
    # )
    # component = partial(gf.c.ring_single_heater, length_x=10)
    # c = add_fiber_array_optical_south_electrical_north(
    #     component=component,
    #     electrical_port_names=["l_e2", "r_e2"],
    #     npads=5,
    # )
    # print(c.name)
    c = add_fiber_array_optical_south_electrical_north()
    c.show(show_ports=False)



from __future__ import annotations

from functools import partial

import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.components.extension import extend_ports
from gdsfactory.components.straight import straight
from gdsfactory.components.taper import taper
from gdsfactory.components.text import text_rectangular
from gdsfactory.typings import (
    ComponentSpec,
    ComponentSpecOrList,
    CrossSectionSpec,
    Float2,
)

edge_coupler_silicon = partial(taper, width2=0.2, length=100, with_two_ports=True)
edge_coupler_silicon_2 = partial(taper, width2=0.2, length=130, with_two_ports=True)


@gf.cell
def edge_coupler_array(
    edge_coupler: ComponentSpec | ComponentSpecOrList = edge_coupler_silicon,
    n: int = 5,
    pitch: float = 127.0,
    x_reflection: bool = False,
    text: ComponentSpec | None = text_rectangular,
    text_offset: Float2 = (10, 20),
    text_rotation: float = 0,
    text_mirror: bool = False,
    angle: float = 0,
    bend: ComponentSpec = bend_euler,
    place_x_by_port: bool = True,
    space_y_by_port: bool = False,
    alignment_port: str = "o1",
) -> Component:
    """Fiber array edge coupler based on an inverse taper.

    Each edge coupler adds a ruler for polishing.

    Args:
        edge_coupler: edge coupler spec or list of edge coupler specs.
        n: number of channels. Overwritten if edge_coupler is a list.
        pitch: Fiber pitch.
        x_reflection: horizontal mirror.
        text: text spec.
        text_offset: from edge coupler.
        text_rotation: text rotation in degrees.
        text_mirror: mirrors the text.
        angle: rotation in degrees.
        bend: bend spec. Used only if angle > 0.
        place_x_by_port: If True it aligns the ports in the x direction.
         If False it aligns the right edge.
        space_y_by_port: If True it spaces the edge couplers based on port.
         If False it just places the y of the edge coupler at the given pitch
        alignment_port: port that we use to align and space edge couplers

    Requires edge coupler waveguide port to face left.

    .. code::
                                          ┌─────────────────┐
                                          │                 │
                                          │                 │
           ───────────────                │                 │
                          ─────────────── │                 │
           o1             edge_coupler_tip│      fiber      │
                          ─────────────── │                 │
           ───────────────                │                 │
                                          │                 │
                                          │                 │
                                          └─────────────────┘

    """
    if not isinstance(edge_coupler, list):
        # Make it a list
        edge_coupler = [edge_coupler] * n
    else:
        n = len(edge_coupler)

    c = Component()
    for i in range(n):
        alias = f"ec_{i}"
        edge_coupler_comp = gf.get_component(edge_coupler[i])
        ref = c.add_ref(edge_coupler_comp, alias=alias)
        ref.rotate(angle)

        if x_reflection:
            ref.mirror()

        if space_y_by_port:
            ref.movey(i * pitch - ref.ports[alignment_port].y)
        else:
            ref.y = i * pitch

        if place_x_by_port:
            ref.movex(-ref.ports[alignment_port].x)
        else:
            ref.xmax = 0

        if angle:
            # straighten the port to a manhattan 180 degree angle to avoid grid errors
            b = c << bend(angle=angle)
            b.connect("o2", ref.ports["o1"])
            c.add_port(f"o{i}", port=b["o1"])

        else:
            for port in ref.get_ports_list():
                c.add_port(f"{port.name}_{i}", port=port)

        if text:
            t = c << gf.get_component(text, text=str(i + 1))
            t.rotate(text_rotation)
            t.move(np.array(text_offset) + (ref.ports[alignment_port].x, i * pitch))
            if text_mirror:
                t.mirror()

    if angle:
        c = c.flatten_offgrid_references()
    c.auto_rename_ports()
    return c


@gf.cell
def edge_coupler_array_with_loopback(
    edge_coupler: ComponentSpec | ComponentSpecOrList = edge_coupler_silicon,
    cross_section: CrossSectionSpec | None = "xs_sc",
    radius: float = 30,
    n: int = 8,
    pitch: float = 127.0,
    extension_length: float = 1.0,
    right_loopback: bool = True,
    x_reflection: bool = False,
    text: ComponentSpec | None = text_rectangular,
    text_offset: Float2 = (0, 0),
    text_rotation: float = 0,
    text_mirror: bool = False,
    bend: ComponentSpec = bend_euler,
    straight: ComponentSpec = straight,
    taper: ComponentSpec | None = None,
    angle: float = 0,
    place_x_by_port: bool = True,
    space_y_by_port: bool = False,
    alignment_port: str = "o1",
) -> Component:
    """Fiber array edge coupler.

    Args:
        edge_coupler: edge coupler.
        cross_section: spec.
        radius: bend radius loopback (um).
        n: number of channels.
        pitch: Fiber pitch (um).
        extension_length: in um.
        right_loopback: adds right loopback.
        x_reflection: horizontal mirror.
        text: Optional text spec.
        text_offset: x, y.
        text_rotation: text rotation in degrees.
        text_mirror: mirrors the text.
        bend: bend spec.
        straight: straight spec.
        taper: taper spec.
        angle: rotation in degrees.
    """
    c = Component()
    ec = edge_coupler_array(
        edge_coupler=edge_coupler,
        n=n,
        pitch=pitch,
        x_reflection=x_reflection,
        text=text,
        text_offset=text_offset,
        text_rotation=text_rotation,
        text_mirror=text_mirror,
        angle=angle,
        bend=bend,
        place_x_by_port=place_x_by_port,
        space_y_by_port=space_y_by_port,
        alignment_port=alignment_port,
    )
    if extension_length > 0:
        ec = extend_ports(
            component=ec,
            port_names=("o1", "o2"),
            length=extension_length,
            extension=partial(
                gf.c.straight, cross_section=cross_section, length=extension_length
            )
            if cross_section
            else straight,
        )

    ec_ref = c << ec
    route1 = gf.routing.get_route(
        ec_ref.ports["o1"],
        ec_ref.ports["o2"],
        bend=bend,
        straight=straight,
        taper=taper,
        cross_section=cross_section,
        radius=radius,
    )
    c.add(route1.references)

    # Override n if it's a list of edge couplers
    if isinstance(edge_coupler, list):
        n = len(edge_coupler)

    if n > 4 and right_loopback:
        route2 = gf.routing.get_route(
            ec_ref.ports[f"o{n-1}"],
            ec_ref.ports[f"o{n}"],
            bend=bend,
            straight=straight,
            taper=taper,
            cross_section=cross_section,
            radius=radius,
        )
        c.add(route2.references)
        for i in range(n - 4):
            c.add_port(str(i), port=ec_ref.ports[f"o{i+3}"])
    elif n > 4:
        for i in range(n - 2):
            c.add_port(str(i), port=ec_ref.ports[f"o{i+3}"])

    c.auto_rename_ports()
    return c


if __name__ == "__main__":
    gf.config.enable_offgrid_ports()
    # c = edge_coupler_array_with_loopback(text_rotation=90)
    # c = edge_coupler_silicon()
    # c = edge_coupler_array(x_reflection=False)
    # c = edge_coupler_array_with_loopback(x_reflection=False)
    # c = edge_coupler_array(angle=8)
    c = edge_coupler_array(
        edge_coupler=[
            edge_coupler_silicon,
            edge_coupler_silicon_2,
            edge_coupler_silicon_2,
            edge_coupler_silicon,
        ],
        angle=8,
        place_x_by_port=True,
        space_y_by_port=True,
        alignment_port="o2",
    )
    # c = edge_coupler_array_with_loopback(
    #     edge_coupler=[
    #         edge_coupler_silicon,
    #         edge_coupler_silicon_2,
    #         edge_coupler_silicon_2,
    #         edge_coupler_silicon,
    #     ],
    #     angle=0,
    #     place_by_port=False,
    # )
    c.show(show_ports=True)



"""CD SEM structures."""
from __future__ import annotations

from functools import partial

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.straight import straight
from gdsfactory.components.text_rectangular import text_rectangular
from gdsfactory.typings import (
    ComponentFactory,
    CrossSectionSpec,
    CrossSectionSpecs,
    Floats,
    Tuple,
)

text_rectangular_mini = partial(text_rectangular, size=1)

widths = 10 * (0.3,)
gaps = 10 * (0.3,)


@cell
def cdsem_straight_density(
    widths: Floats = widths,
    gaps: Floats = gaps,
    length: float = 420.0,
    label: str = "",
    cross_section: CrossSectionSpec | CrossSectionSpecs = "xs_sc",
    text: ComponentFactory | None = text_rectangular_mini,
) -> Component:
    """Returns sweep of dense straight lines.

    Args:
        widths: list of widths.
        gaps: list of gaps.
        length: of the lines.
        label: defaults to widths[0] gaps[0].
        cross_section: spec. Can be a list and then each line has a corresponding cross_section.
        text: optional function for text.
    """
    c = Component()
    label = label or f"{int(widths[0]*1e3)} {int(gaps[0]*1e3)}"

    if isinstance(cross_section, Tuple):
        if len(cross_section) != len(widths):
            raise ValueError(
                "The number of specified cross sections does not correspond to the number of widths"
            )
    else:
        cross_section = [cross_section] * len(widths)

    ymin = 0
    for width, gap, xs in zip(widths, gaps, cross_section):
        tooth_ref = c << straight(length=length, cross_section=xs, width=width)
        tooth_ref.ymin = ymin
        ymin += width + gap

    if text:
        marker_label = c << gf.get_component(text, text=f"{label}")
        marker_label.xmin = tooth_ref.xmax + 5
    return c


if __name__ == "__main__":
    c = cdsem_straight_density(
        widths=(0.2, 0.3), gaps=(0.1, 0.2), cross_section="xs_sc"
    )
    c.show(show_ports=True)



from __future__ import annotations

import warnings
from functools import partial

import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.compass import compass
from gdsfactory.components.via import via1, via2, viac
from gdsfactory.components.wire import wire_corner45
from gdsfactory.typings import (
    Callable,
    ComponentFactory,
    ComponentSpec,
    Float2,
    Floats,
    LayerSpec,
    LayerSpecs,
)


@gf.cell
def via_stack(
    size=(11.0, 11.0),
    layers: tuple[LayerSpec | None, ...] = ("M1", "M2", "MTOP"),
    layer_offsets: Floats | None = None,
    vias: tuple[ComponentSpec | None, ...] | None = (via1, via2, None),
    layer_port: LayerSpec | None = None,
    correct_size: bool = True,
    slot_horizontal: bool = False,
    slot_vertical: bool = False,
    post_process: Callable | None = None,
) -> Component:
    """Rectangular via array stack.

    You can use it to connect different metal layers or metals to silicon.
    You can use the naming convention via_stack_layerSource_layerDestination
    contains 4 ports (e1, e2, e3, e4)

    also know as Via array
    http://www.vlsi-expert.com/2017/12/vias.html

    spacing = via.info['spacing']
    enclosure = via.info['enclosure']

    Args:
        size: of the layers.
        layers: layers on which to draw rectangles.
        layer_offsets: Optional offsets for each layer with respect to size.
            positive grows, negative shrinks the size.
        vias: vias to use to fill the rectangles.
        layer_port: if None assumes port is on the last layer.
        correct_size: if True, if the specified dimensions are too small it increases
            them to the minimum possible to fit a via.
        slot_horizontal: if True, then vias are horizontal.
        slot_vertical: if True, then vias are vertical.
    """
    width_m, height_m = size
    a = width_m / 2
    b = height_m / 2

    layers = layers or []
    layer_offsets = layer_offsets or [0] * len(layers)

    elements = {len(layers), len(layer_offsets), len(vias)}
    if len(elements) > 1:
        warnings.warn(
            f"Got {len(layers)} layers, {len(layer_offsets)} layer_offsets, {len(vias)} vias",
            stacklevel=3,
        )

    if layers:
        layer_port = layer_port or layers[-1]

    c = Component()
    c.height = height_m
    c.info["size"] = tuple(size)
    c.info["xsize"] = size[0]
    c.info["ysize"] = size[1]
    if layer_port:
        c.info["layer"] = layer_port

    for layer, offset in zip(layers, layer_offsets):
        size_m = (width_m + 2 * offset, height_m + 2 * offset)
        if layer and layer == layer_port:
            ref = c << compass(size=size_m, layer=layer, port_type="electrical")
            c.add_ports(ref.ports)
        elif layer is not None:
            ref = c << compass(size=size_m, layer=layer, port_type="electrical")

    vias = vias or []
    for via, offset in zip(vias, layer_offsets):
        if via is not None:
            width, height = size
            width += 2 * offset
            height += 2 * offset
            _via = gf.get_component(via)
            w, h = _via.info["xsize"], _via.info["ysize"]
            enclosure = _via.info["enclosure"]
            pitch_x, pitch_y = _via.info["xspacing"], _via.info["yspacing"]

            if slot_horizontal:
                width = size[0] - 2 * enclosure
                via = gf.get_component(via, size=(width, h))
                nb_vias_x = 1
                nb_vias_y = abs(height - h - 2 * enclosure) / pitch_y + 1

            elif slot_vertical:
                height = size[1] - 2 * enclosure
                via = gf.get_component(via, size=(w, height))
                nb_vias_x = abs(width - w - 2 * enclosure) / pitch_x + 1
                nb_vias_y = 1
            else:
                via = _via
                nb_vias_x = abs(width - w - 2 * enclosure) / pitch_x + 1
                nb_vias_y = abs(height - h - 2 * enclosure) / pitch_y + 1

            min_width = w + enclosure
            min_height = h + enclosure

            if (
                min_width > width
                and correct_size
                or min_width <= width
                and min_height > height
                and correct_size
            ):
                warnings.warn(
                    f"Changing size from ({width}, {height}) to ({min_width}, {min_height}) to fit a via!",
                    stacklevel=3,
                )
                width = max(min_width, width)
                height = max(min_height, height)
            elif min_width > width or min_height > height:
                raise ValueError(f"size {size} is too small to fit a {(w, h)} um via")

            nb_vias_x = int(np.floor(nb_vias_x)) or 1
            nb_vias_y = int(np.floor(nb_vias_y)) or 1
            ref = c.add_array(
                via, columns=nb_vias_x, rows=nb_vias_y, spacing=(pitch_x, pitch_y)
            )
            if ref.xsize + enclosure > width or ref.ysize + enclosure > height:
                warnings.warn(
                    f"size = {size} for layer {layer} violates min enclosure"
                    f" {enclosure} for via {via.name!r}",
                    stacklevel=3,
                )

            a = width / 2
            b = height / 2
            cw = (width - (nb_vias_x - 1) * pitch_x - w) / 2
            ch = (height - (nb_vias_y - 1) * pitch_y - h) / 2
            x0 = -a + cw + w / 2
            y0 = -b + ch + h / 2
            ref.move((x0, y0))

    if post_process:
        post_process(c)

    return c


@gf.cell
def via_stack_corner45(
    width: float = 10,
    layers: tuple[LayerSpec | None, ...] = ("M1", "M2", "MTOP"),
    layer_offsets: Floats | None = None,
    vias: tuple[ComponentSpec | None, ...] | None = (via1, via2, None),
    layer_port: LayerSpec | None = None,
    correct_size: bool = True,
) -> Component:
    """Rectangular via array stack at a 45 degree angle.

    spacing = via.info['spacing']
    enclosure = via.info['enclosure']

    Args:
        width: of the corner45.
        layers: layers on which to draw rectangles.
        layer_offsets: Optional offsets for each layer with respect to size.
            positive grows, negative shrinks the size.
        vias: vias to use to fill the rectangles.
        layer_port: if None assumes port is on the last layer.
        correct_size: if True, if the specified dimensions are too small it increases
            them to the minimum possible to fit a via.
    """

    height = width
    layers = layers or []
    layer_offsets = layer_offsets or [0] * len(layers)

    elements = {len(layers), len(layer_offsets), len(vias)}
    if len(elements) > 1:
        warnings.warn(
            f"Got {len(layers)} layers, {len(layer_offsets)} layer_offsets, {len(vias)} vias",
            stacklevel=3,
        )

    if layers:
        layer_port = layer_port or layers[-1]

    c = Component()
    if layer_port:
        c.info["layer"] = layer_port

    for layer, offset in zip(layers, layer_offsets):
        if layer and layer == layer_port:
            ref = c << wire_corner45(
                width=width + 2 * offset, layer=layer, with_corner90_ports=False
            )
            c.add_ports(ref.ports)
        elif layer is not None:
            ref = c << wire_corner45(
                width=width + 2 * offset, layer=layer, with_corner90_ports=False
            )

    width_corner = width
    width, height = ref.size
    xmin = ref.xmin
    ymin = ref.ymin

    vias = vias or []
    for via, offset in zip(vias, layer_offsets):
        if via is not None:
            width45 = (
                2 * (width_corner + 2 * offset) * np.cos(np.deg2rad(45))
            )  # Width in the x direction
            _via = gf.get_component(via)
            w, h = _via.info["xsize"], _via.info["ysize"]
            enclosure = _via.info["enclosure"]
            pitch_x, pitch_y = _via.info["xspacing"], _via.info["yspacing"]

            via = _via

            min_width = w + enclosure
            min_height = h + enclosure

            if (
                min_width > width45
                and correct_size
                or min_width <= width45
                and min_height > height
                and correct_size
            ):
                warnings.warn(
                    f"Changing size from ({width}, {height}) to ({min_width}, {min_height}) to fit a via!",
                    stacklevel=3,
                )
                width45 = max(min_width, width45)
                height = max(min_height, height)
            elif min_width > width45 or min_height > height:
                raise ValueError(
                    f"{min_width=} > {width=} or {min_height=} > {height=}"
                )

            # Keep placing rows until we cover the whole height
            y_covered = enclosure

            while y_covered + enclosure < height:
                y = ymin + y_covered + h / 2  # Position of the via

                # x offset from the edge of the metal to make sure enclosure is fulfilled
                xoff_enc = 2 * enclosure * np.cos(np.deg2rad(45))
                xoff = (y_covered + h) * np.tan(np.deg2rad(45)) + xoff_enc

                xpos0 = xmin + xoff

                # Calculate the number of vias that fit in a given width
                if (y_covered + h) < (height - width45):
                    # The x width is width45
                    xwidth = width45
                else:
                    # The x width is decreasing
                    xwidth = (height - (y_covered + h)) * np.tan(np.deg2rad(45))

                if min_width > xwidth:
                    pass
                else:
                    vias_per_row = (
                        xwidth - 2 * xoff_enc - 2 * h * np.tan(np.deg2rad(45))
                    ) / (pitch_x) + 1
                    # Place the vias at the given x, y
                    for i in range(int(vias_per_row)):
                        ref = c << via
                        ref.center = (xpos0 + pitch_x * i + w / 2, y)

                y_covered = y_covered + h + pitch_y

    return c


@gf.cell
def via_stack_corner45_extended(
    corner: ComponentSpec = via_stack_corner45,
    via_stack: ComponentSpec = via_stack,
    width: float = 3,
    length: float = 10,
) -> Component:
    """Rectangular via array stack at a 45 degree angle.

    Args:
        corner: corner component.
        straight: straight component.
        width: of the corner45.
        length: of the straight.
    """
    c = gf.Component()
    corner = c << gf.get_component(corner, width=width / np.sqrt(2))
    s = gf.get_component(via_stack, size=(length, width))
    sr = c << s
    sl = c << s
    sr.connect("e1", corner.ports["e1"])
    sl.connect("e1", corner.ports["e2"])
    return c


@gf.cell
def via_stack_from_rules(
    size: Float2 = (1.2, 1.2),
    layers: LayerSpecs = ("M1", "M2", "MTOP"),
    layer_offsets: tuple[float, ...] | None = None,
    vias: tuple[ComponentSpec | None, ...] | None = (via1, via2),
    via_min_size: tuple[Float2, ...] = ((0.2, 0.2), (0.2, 0.2)),
    via_min_gap: tuple[Float2, ...] = ((0.1, 0.1), (0.1, 0.1)),
    via_min_enclosure: Float2 = (0.15, 0.25),
    layer_port: LayerSpec | None = None,
) -> Component:
    """Rectangular via array stack, with optimized dimension for vias.

    Uses inclusion, minimum width, and minimum spacing rules to place the maximum number of individual vias,
    each with maximum via area.

    Args:
        size: of the layers, len(size).
        layers: layers on which to draw rectangles.
        layer_offsets: Optional offsets for each layer with respect to size.
            positive grows, negative shrinks the size.
        vias: list of base via components to modify.
        via_min_size: via minimum x, y dimensions.
        via_min_gap: via minimum x, y distances.
        via_min_enclosure: via minimum inclusion into connecting layers.
        layer_port: if None assumes port is on the last layer.
    """
    width, height = size
    a = width / 2
    b = height / 2

    layers = layers or []

    if layers:
        layer_port = layer_port or layers[-1]

    c = Component()
    c.height = height
    c.info["xsize"] = size[0]
    c.info["ysize"] = size[1]
    c.info["layer"] = layer_port

    layer_offsets = layer_offsets or [0] * len(layers)

    for layer, offset in zip(layers, layer_offsets):
        size = (width + 2 * offset, height + 2 * offset)
        if layer and layer == layer_port:
            ref = c << compass(size=size, layer=layer, port_type="electrical")
            c.add_ports(ref.ports)
        elif layer:
            ref = c << compass(size=size, layer=layer, port_type="electrical")

    vias = vias or []
    for current_via, min_size, min_gap, min_enclosure in zip(
        vias, via_min_size, via_min_gap, via_min_enclosure
    ):
        if current_via is not None:
            # Optimize via
            via = gf.get_component(
                optimized_via(current_via, size, min_size, min_gap, min_enclosure)
            )

            w, h = via.info["size"]
            g = via.info["enclosure"]
            pitch_x, pitch_y = via.info["spacing"]

            nb_vias_x = (width - w - 2 * g) / pitch_x + 1
            nb_vias_y = (height - h - 2 * g) / pitch_y + 1

            nb_vias_x = int(np.floor(nb_vias_x)) or 1
            nb_vias_y = int(np.floor(nb_vias_y)) or 1
            ref = c.add_array(
                via, columns=nb_vias_x, rows=nb_vias_y, spacing=(pitch_x, pitch_y)
            )

            cw = (width - (nb_vias_x - 1) * pitch_x - w) / 2
            ch = (height - (nb_vias_y - 1) * pitch_y - h) / 2
            x0 = -a + cw + w / 2
            y0 = -b + ch + h / 2
            ref.move((x0, y0))

    return c


def optimized_via(
    base_via: ComponentSpec = "VIAC",
    size: tuple[float, float] = (11.0, 11.0),
    min_via_size: tuple[float, float] = (0.3, 0.3),
    min_via_gap: tuple[float, float] = (0.1, 0.1),
    min_via_enclosure: float = 0.2,
) -> ComponentFactory:
    """Given a target total inclusion size, returns an optimized dimension for the via.

    Uses inclusion, minimum width, and minimum spacing rules to place the maximum number of individual vias, with maximum via area.

    Arguments:
        base_via: to modify.
        size: of the target enclosing medium.
        min_via_size: minimum size the vias can take.
        min_via_gap: minimum distance between vias.
        min_via_enclosure: minimum distance between edge of enclosing medium and nearest via edge.
    """
    via_size = [0, 0]
    for dim in [0, 1]:
        via_area = size[dim] - 2 * min_via_enclosure + min_via_gap[dim]
        num_vias = int(via_area / (min_via_size[dim] + min_via_gap[dim]))
        try:
            via_size[dim] = float(via_area / num_vias) - min_via_gap[dim]
        except ZeroDivisionError as e:
            raise RuntimeError(
                "Cannot fit vias with specified minimum dimensions in provided space."
            ) from e

    return partial(
        base_via,
        size=via_size,
        gap=min_via_gap,
        spacing=None,
        enclosure=min_via_enclosure,
    )


via_stack_m1_m3 = partial(
    via_stack,
    layers=("M1", "M2", "MTOP"),
    vias=(via1, via2, None),
)
via_stack_m2_m3 = partial(
    via_stack,
    layers=("M2", "MTOP"),
    vias=(via2, None),
)

via_stack_slab_m3 = partial(
    via_stack,
    layers=("SLAB90", "M1", "M2", "MTOP"),
    vias=(viac, via1, via2, None),
)
via_stack_slab_m2 = partial(
    via_stack,
    layers=("SLAB90", "M1", "M2"),
    vias=(viac, via1, None),
)
via_stack_npp_m1 = partial(
    via_stack,
    layers=("WG", "NPP", "M1"),
    vias=(None, None, viac),
)
via_stack_slab_npp_m3 = partial(
    via_stack,
    layers=("SLAB90", "NPP", "M1"),
    vias=(None, None, viac),
)
via_stack_heater_m2 = partial(via_stack, layers=("HEATER", "M2"), vias=(None, via1))
via_stack_heater_mtop = via_stack_heater_m3 = partial(
    via_stack, layers=("HEATER", "M2", "MTOP"), vias=(None, via1, via2)
)


if __name__ == "__main__":
    # c = via_stack()
    # c = gf.pack([via_stack_slab_m3, via_stack_heater_mtop])[0]
    # c = via_stack_slab_m3(size=(100, 10), slot_vertical=True)
    # c = via_stack_from_rules()
    c = via_stack_corner45()
    # c = via_stack_corner45_extended()
    c.show(show_ports=True)
